#!/usr/bin/env python3
# -*- coding: utf-8 -*-
################################################################################
#    Wisteria Copyright (C) 2021 suizokukan
#    Contact: suizokukan _A.T._ orange dot fr
#
#    This file is part of Wisteria.
#    Wisteria is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Wisteria is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Wisteria.  If not, see <http://www.gnu.org/licenses/>.
################################################################################
"""
    Wisteria project : wisteria/cwc/simple/works_as_expected.py

    initialize() and works_as_expected() functions for all cwc/simple/ classes.


    ___________________________________________________________________________

    o  initialize(obj)
    o  works_as_expected(obj)
"""


def initialize(obj):
    """
        initialize() function

        _______________________________________________________________________

        ARGUMENT: <obj>, the object to be initialized

        RETURNED OBJECT: <obj>, the now initialized object
    """
    return obj


def works_as_expected(data_name,
                      obj=None):
    """
        works_as_expected()

        works_as_expected() function for basic types defined DATA/UNAVAILABLE_DATA.


        _______________________________________________________________________

        (pimydoc)works_as_expected arguments and returned value
        ⋅All works_as_expected() functions are supposed to (1) say if <data_name> is in
        ⋅the scope of this function (2) and say if <obj> works as expected.
        ⋅
        ⋅ARGUMENTS:
        ⋅    o  data_name:   (str)data_name of the <obj>ect
        ⋅    o  obj:         (None or any object) object to be checked
        ⋅
        ⋅RETURNED VALUE:
        ⋅    (<obj> is None)     (bool)<data_name> is known
        ⋅    (<obj> is not None) <obj> works as expected.
    """
    if obj is None:
        # for CWC objects, this case is never reached.
        return True

    return obj.integer == 3 and \
        obj.string == "string" and \
        obj._list == [1, 2, 3] and \
        obj._dict == {"key1": "value1"}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
################################################################################
#    Wisteria Copyright (C) 2021 suizokukan
#    Contact: suizokukan _A.T._ orange dot fr
#
#    This file is part of Wisteria.
#    Wisteria is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Wisteria is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Wisteria.  If not, see <http://www.gnu.org/licenses/>.
################################################################################
"""
    Wisteria project : wisteria/wisteria/cwc/simple/s_iaswn.py

    Default cwc/SimpleClass class, a simple class made of the following types:
    - int
    - str
    - list
    - dict
        Note that dicts have only non string keys.

    ___________________________________________________________________________

    o  SimpleClass class
"""
from iaswn.iaswn import Iaswn


class SimpleClass(Iaswn):
    def __eq__(self,
               other):
        """SimpleClass.__eq__()"""
        return self.integer == other.integer and \
            self.string == other.string and \
            self._list == other._list and \
            self._dict == other._dict

    def __init__(self,
                 integer=3,
                 string="string",
                 _list=[1, 2, 3],
                 _dict={"key1": "value1"}):
        """SimpleClass.__init__()"""
        Iaswn.__init__(self)
        self.integer = integer
        self.string = string
        self._list = _list
        self._dict = _dict



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
################################################################################
#    Wisteria Copyright (C) 2021 suizokukan
#    Contact: suizokukan _A.T._ orange dot fr
#
#    This file is part of Wisteria.
#    Wisteria is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Wisteria is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Wisteria.  If not, see <http://www.gnu.org/licenses/>.
################################################################################
"""
    Wisteria project : wisteria/results.py

    Fonctions to compute results, especially compute_results().

    ___________________________________________________________________________

    o  get_serializers_selection(serializer1, serializer2)
    o  get_data_selection(cmpdata, config)
    o  compute_results(config, serializer1, serializer2, cmpdata)
    o  init_planned_transcodings(serializer1, serializer2, cmpdata, config)
"""
from rich.console import Console
from rich.progress_bar import ProgressBar

import wisteria.data
import wisteria.globs
from wisteria.globs import VERBOSITY_NORMAL, VERBOSITY_DEBUG, VERBOSITY_DETAILS
from wisteria.globs import PROGRESSBAR_LENGTH
from wisteria.wisteriaerror import WisteriaError
from wisteria.msg import msgdebug, msginfo, msgerror
from wisteria.serializers_classes import SerializationResults
from wisteria.utils import strdigest
from wisteria.cwc.cwc_utils import is_a_cwc_name, moduleininame_to_modulefullrealname
from wisteria.cwc.cwc_utils import modulefullrealname_to_modulerealname
from wisteria.cwc.cwc_utils import modulefullrealname_to_classname
from wisteria.cwc.cwc_utils import is_this_an_appropriate_module_for_serializer
from wisteria.cwc.cwc_utils import modulefullrealname_to_waemodulename


def get_serializers_selection(serializer1,
                              serializer2):
    """
        get_serializers_selection()

        Return a tuple of all serializers defined by (str)<serializer1>, (str)<serializer2>.

        _______________________________________________________________________

        ARGUMENTS:
        o    <serializer1> : value returned by read_cmpstring()
        o    <serializer2> : value returned by read_cmpstring()

        RETURNED VALUE: a tuple of str, i.e. all serializers defined by (str)<serializer1>,
                        (str)<serializer2>
    """
    res = set()

    if serializer1 == 'all':
        for serializer in wisteria.globs.SERIALIZERS:
            res.add(serializer)
    else:
        res.add(serializer1)

    if serializer2 == 'all':
        for serializer in wisteria.globs.SERIALIZERS:
            res.add(serializer)
    else:
        res.add(serializer2)

    return tuple(res)


def get_data_selection(cmpdata,
                       config):
    """
        get_data_selection()

        Return a tuple of the data objects names selected by the <config>.

        _______________________________________________________________________

        ARGUMENTS:
        o  cmpdata: (str)'all', 'ini', 'cwc' or 'allbutcwc'
        o  config:  (dict) dict returned by read_cfgfile()

        RETURNED VALUE: (tuple) a tuple of (str)data keys
    """
    res = []

    if cmpdata == 'all':
        res = tuple(wisteria.globs.DATA.keys())
    elif cmpdata == 'ini':
        if config["data selection"]["data selection"] == 'all':
            res = tuple(config["objects"].keys())
        elif config["data selection"]["data selection"] == 'only if yes':
            res = tuple(data_name for data_name in config["data objects"]
                        if config["data objects"][data_name])
        elif config["data selection"]["data selection"].startswith("data set/"):
            res = tuple(config["data sets"][config["data selection"]["data selection"]])
        else:
            raise WisteriaError(
                "(ERRORID020) "
                "Can't understand a value given to \\[data selection]'data selection': "
                f"what is '{config['data selection']['data selection']}' ? "
                "Known values are 'all', 'ini' and 'data set/xxx' "
                "where xxx is a string.")
    elif cmpdata == 'cwc':
        res = tuple(data_name for data_name in wisteria.globs.DATA
                    if is_a_cwc_name(data_name))
    elif cmpdata == 'allbutcwc':
        res = tuple(data_name for data_name in wisteria.globs.DATA
                    if not is_a_cwc_name(data_name))
    else:
        raise WisteriaError(
            "(ERRORID021) Can't understand a value given to the 'data' part in --cmp: "
            f"what is '{cmpdata}' ? "
            "Known values are 'all', 'cwc' and 'ini'. ")

    return res


def compute_results():
    """
        compute_results()

        Create a SerializationResults object and try to fill it with all
        required encodings/decodings defined by PLANNED_TRANSCODINGS
        _______________________________________________________________________

        RETURNED VALUE:    (SerializationResults, None) if no error occured
                        or (None, (int)exit_code) if an error occured
    """
    def erase_progress_bar():
        """
            erase_progress_bar()

            Erase the progress bar so that it is possible to write text over
            the ancient progress bar.

            Show the cursor.

            (pimydoc)progress bar
            ⋅A progress bar is displayed only if verbosity is set to 1 (normal).
            ⋅If verbosity is set to 0 (minimal), the progress bar is hidden since no
            ⋅console output is authorized: it's important for scripts calling the
            ⋅project from the outside.
            ⋅If verbosity is set to 2 (details) or 3 (debug), the progress bar is hidden
            ⋅in order to avoid mixing the progress bar with the text displayed while
            ⋅computing the result, which is unpleasant to see.
        """
        if wisteria.globs.ARGS.verbosity == VERBOSITY_NORMAL:
            # the following lines make the progress bar disappear.
            # the next rprint() will overwrite the spaces that are about
            # to be added:
            if PROGRESSBAR_LENGTH is None:
                console.file.write(" "*console.width)
            else:
                console.file.write(" "*PROGRESSBAR_LENGTH)
            console.file.write("\r")

            console.show_cursor(True)

    try:
        results = SerializationResults()

        planned_transcodings_number = len(wisteria.globs.PLANNED_TRANSCODINGS)

        # (pimydoc)progress bar
        # ⋅A progress bar is displayed only if verbosity is set to 1 (normal).
        # ⋅If verbosity is set to 0 (minimal), the progress bar is hidden since no
        # ⋅console output is authorized: it's important for scripts calling the
        # ⋅project from the outside.
        # ⋅If verbosity is set to 2 (details) or 3 (debug), the progress bar is hidden
        # ⋅in order to avoid mixing the progress bar with the text displayed while
        # ⋅computing the result, which is unpleasant to see.
        if wisteria.globs.ARGS.verbosity == VERBOSITY_NORMAL:
            console = Console()
            progressbar = ProgressBar(width=PROGRESSBAR_LENGTH,
                                      total=planned_transcodings_number)
            console.show_cursor(False)
            progressbar_index = 0

        # ---- real work ------------------------------------------------------
        if wisteria.globs.ARGS.verbosity >= VERBOSITY_DETAILS:
            msginfo("Please wait until all required encodings/decodings have been computed.")

        # (pimydoc)PLANNED_TRANSCODINGS
        # ⋅list of str:
        # ⋅    (str)serializer, (str)data_name, (str)fingerprint
        # ⋅
        # ⋅Initialized by results.py:init_planned_transcodings()
        for (transcoding_index,
             (serializer,
              data_name,
              fingerprint)) in enumerate(wisteria.globs.PLANNED_TRANSCODINGS):

            if serializer not in results:
                results[serializer] = {}

            if not is_a_cwc_name(data_name):
                # ==== <data_name> is NOT A CWC CLASS =========================
                if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
                    msgdebug(f"({transcoding_index+1}/{planned_transcodings_number}) "
                             "About to call transcoding functions "
                             f"for serializer='{serializer}' "
                             f"and data name='{data_name}' "
                             f"[{fingerprint}]")

                results[serializer][data_name] = wisteria.globs.SERIALIZERS[serializer].func(
                    action="serialize",
                    obj=wisteria.globs.DATA[data_name],
                    obj_data_name=data_name,
                    fingerprint=fingerprint,
                    works_as_expected=wisteria.data.works_as_expected
                    if wisteria.data.works_as_expected(data_name=data_name,
                                                       obj=None) is True else None)
            else:
                # ==== <data_name> is a CWC CLASS =============================
                # data_name: e.g. "cwc.pgnreader.cwc_default.chessgames"
                #                > "cwc.pgnreader.cwc_default.ChessGames"
                data_name = moduleininame_to_modulefullrealname(data_name)
                # data_name__strmodule: e.g. "cwc.pgnreader.cwc_default"
                data_name__strmodule = modulefullrealname_to_modulerealname(data_name)
                # data_name__strmodule_wae: e.g. "cwc.pgnreader.works_as_expected"
                data_name__strwaemodulename = modulefullrealname_to_waemodulename(data_name)
                # data_name__strclassname: e.g. "ChessGames"
                data_name__strclassname = modulefullrealname_to_classname(data_name)

                if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
                    msgdebug("About to call serialize/unserialize function "
                             f"for serializer='{serializer}' "
                             f"and (cwc) data name='{data_name}' "
                             f"[{fingerprint}]")

                cwc_object = \
                    getattr(wisteria.globs.MODULES[data_name__strwaemodulename],
                            "initialize")(getattr(
                                wisteria.globs.MODULES[data_name__strmodule],
                                data_name__strclassname)())

                results[serializer][data_name] = \
                    wisteria.globs.SERIALIZERS[serializer].func(
                        action="serialize",
                        obj=cwc_object,
                        obj_data_name=data_name,
                        fingerprint=fingerprint,
                        works_as_expected=getattr(
                            wisteria.globs.MODULES[data_name__strwaemodulename],
                            "works_as_expected"))

            # (pimydoc)progress bar
            # ⋅A progress bar is displayed only if verbosity is set to 1 (normal).
            # ⋅If verbosity is set to 0 (minimal), the progress bar is hidden since no
            # ⋅console output is authorized: it's important for scripts calling the
            # ⋅project from the outside.
            # ⋅If verbosity is set to 2 (details) or 3 (debug), the progress bar is hidden
            # ⋅in order to avoid mixing the progress bar with the text displayed while
            # ⋅computing the result, which is unpleasant to see.
            if wisteria.globs.ARGS.verbosity == VERBOSITY_NORMAL:
                progressbar_index += 1
                progressbar.update(progressbar_index)
                console.print(progressbar)
                console.file.write("\r")

            if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
                # is serializer/data_name has been skipped, no results[serializer][data_name],
                # hence the following "if" statement:
                if data_name in results[serializer]:
                    msgdebug(f"result: {results[serializer][data_name]} "
                             f"[{fingerprint}]")

        # (pimydoc)progress bar
        # ⋅A progress bar is displayed only if verbosity is set to 1 (normal).
        # ⋅If verbosity is set to 0 (minimal), the progress bar is hidden since no
        # ⋅console output is authorized: it's important for scripts calling the
        # ⋅project from the outside.
        # ⋅If verbosity is set to 2 (details) or 3 (debug), the progress bar is hidden
        # ⋅in order to avoid mixing the progress bar with the text displayed while
        # ⋅computing the result, which is unpleasant to see.
        erase_progress_bar()

        if not results.finish_initialization():
            msgerror("(ERRORID015) Incorrect data, the program has to stop.")

            # (pimydoc)exit codes
            # ⋅*  0: normal exit code
            # ⋅*  1: normal exit code after --checkup
            # ⋅*  2: normal exit code after --downloadconfigfile
            # ⋅*  3: normal exit code after --mymachine
            # ⋅*  4: normal exit code (no data to handle)
            # ⋅* -1: error, given config file can't be read (missing or ill-formed file)
            # ⋅* -2: error, ill-formed --cmp string
            # ⋅* -3: error, ill-formed --output string
            # ⋅* -4: error, missing required module
            # ⋅* -8: error, STR2REPORTSECTION_KEYS and STR2REPORTSECTION don't match
            # ⋅* -100: internal error, data can't be loaded
            # ⋅* -101: internal error, an error occured while computing the results
            # ⋅* -102: internal error, an error occured in main()
            # ⋅* -103: internal error, can't initialize PLANNED_TRANSCODINGS
            return None, -100
        if results.dataobjs_number == 0:
            msginfo("No data to handle, the program can stop.")

            # (pimydoc)exit codes
            # ⋅*  0: normal exit code
            # ⋅*  1: normal exit code after --checkup
            # ⋅*  2: normal exit code after --downloadconfigfile
            # ⋅*  3: normal exit code after --mymachine
            # ⋅*  4: normal exit code (no data to handle)
            # ⋅* -1: error, given config file can't be read (missing or ill-formed file)
            # ⋅* -2: error, ill-formed --cmp string
            # ⋅* -3: error, ill-formed --output string
            # ⋅* -4: error, missing required module
            # ⋅* -8: error, STR2REPORTSECTION_KEYS and STR2REPORTSECTION don't match
            # ⋅* -100: internal error, data can't be loaded
            # ⋅* -101: internal error, an error occured while computing the results
            # ⋅* -102: internal error, an error occured in main()
            # ⋅* -103: internal error, can't initialize PLANNED_TRANSCODINGS
            return None, 4

        return results, None

    except WisteriaError as exception:
        msgerror(f"An error occured: {exception}")

        # (pimydoc)exit codes
        # ⋅*  0: normal exit code
        # ⋅*  1: normal exit code after --checkup
        # ⋅*  2: normal exit code after --downloadconfigfile
        # ⋅*  3: normal exit code after --mymachine
        # ⋅*  4: normal exit code (no data to handle)
        # ⋅* -1: error, given config file can't be read (missing or ill-formed file)
        # ⋅* -2: error, ill-formed --cmp string
        # ⋅* -3: error, ill-formed --output string
        # ⋅* -4: error, missing required module
        # ⋅* -8: error, STR2REPORTSECTION_KEYS and STR2REPORTSECTION don't match
        # ⋅* -100: internal error, data can't be loaded
        # ⋅* -101: internal error, an error occured while computing the results
        # ⋅* -102: internal error, an error occured in main()
        # ⋅* -103: internal error, can't initialize PLANNED_TRANSCODINGS
        return None, -101


def init_planned_transcodings(serializer1,
                              serializer2,
                              cmpdata,
                              config):
    """
        init_planned_transcodings()

        Initialize wisteria.globs.PLANNED_TRANSCODINGS.

        (pimydoc)PLANNED_TRANSCODINGS
        ⋅list of str:
        ⋅    (str)serializer, (str)data_name, (str)fingerprint
        ⋅
        ⋅Initialized by results.py:init_planned_transcodings()


        _______________________________________________________________________

        ARGUMENTS:
        o  config               : None if no config file, otherwise the dict
                                  returned by read_cfgfile().
        o  (str)serializer1
        o  (str)serializer2
        o  (str)cmpdata         : 'all', 'ini', 'cwc' or 'allbutcwc'

        RETURNED VALUE: (bool)success
    """
    try:
        wisteria.globs.PLANNED_TRANSCODINGS = []

        # serializers and data to be used through the tests:
        serializers = get_serializers_selection(serializer1, serializer2)
        if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
            msgdebug(f"serializers to be used are: {serializers}")
        dataobjs = get_data_selection(cmpdata, config)
        if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
            msgdebug(f"data objs to be used are: {dataobjs}")

        for serializer in sorted(serializers):
            for dataobj in sorted(dataobjs):
                fingerprint = strdigest(serializer+dataobj)

                if not is_a_cwc_name(dataobj):
                    # (pimydoc)PLANNED_TRANSCODINGS
                    # ⋅list of str:
                    # ⋅    (str)serializer, (str)data_name, (str)fingerprint
                    # ⋅
                    # ⋅Initialized by results.py:init_planned_transcodings()
                    wisteria.globs.PLANNED_TRANSCODINGS.append((serializer,
                                                                dataobj,
                                                                fingerprint))
                else:
                    # data_name: e.g. "cwc.pgnreader.cwc_default.chessgames"
                    #                > "cwc.pgnreader.cwc_default.ChessGames"
                    data_name = moduleininame_to_modulefullrealname(dataobj)
                    # data_name__strmodule: e.g. "cwc.pgnreader.cwc_default"
                    data_name__strmodule = modulefullrealname_to_modulerealname(data_name)

                    # e.g. if data_name__strmodule is "cwc.pgnreader.cwc_default" and
                    #      if SERIALIZERS[serializer].cwc is "default" > True
                    #
                    # e.g. if data_name__strmodule is "cwc.pgnreader.cwc_default" and
                    #      if SERIALIZERS[serializer].cwc is "iaswn" > False - we skip.
                    if not is_this_an_appropriate_module_for_serializer(data_name__strmodule,
                                                                        serializer):
                        # we have to skip {serializer, data_name}.
                        if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
                            msgdebug(
                                f"We skip {serializer=} / {data_name=}, "
                                f"since {data_name=} is a cwc name and "
                                f"since data name module '{data_name__strmodule}' "
                                f"doesn't end with SERIALIZERS['{serializer}'].cwc="
                                f"'{wisteria.globs.SERIALIZERS[serializer].cwc}' .")
                    else:
                        # (pimydoc)PLANNED_TRANSCODINGS
                        # ⋅list of str:
                        # ⋅    (str)serializer, (str)data_name, (str)fingerprint
                        # ⋅
                        # ⋅Initialized by results.py:init_planned_transcodings()
                        wisteria.globs.PLANNED_TRANSCODINGS.append((serializer,
                                                                    dataobj,
                                                                    fingerprint))

    except WisteriaError as exception:
        msgerror(f"(ERRORID041) Can't set PLANNED_TRANSCODINGS since an error occured: {exception}")
        return False

    return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
################################################################################
#    Wisteria Copyright (C) 2021 suizokukan
#    Contact: suizokukan _A.T._ orange dot fr
#
#    This file is part of Wisteria.
#    Wisteria is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Wisteria is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Wisteria.  If not, see <http://www.gnu.org/licenses/>.
################################################################################
"""
    Wisteria project : wisteria/globs.py

    Global definitions.

    ___________________________________________________________________________

    o  ARGS

    o  CWC_MODULES

    o  DATA
    o  UNAVAILABLE_DATA

    o  DEBUG_CONSOLEWIDTH

    o  DEFAULT_CONFIGFILE_NAME
    o  DEFAULT_CONFIGFILE_URL

    o  DEFAULT_EXPORTREPORT_FILENAME

    o  DEFAULT_REPORTFILE_NAME

    o  FILECONSOLE
    o  FILECONSOLE_FILEOBJECT

    o  MODULES

    o  OUTPUT

    o  PLANNED_TRANSCODINGS

    o  PLATFORM_SYSTEM

    o  PROGRESSBAR_LENGTH

    o  REGEX_CMP
    o  REGEX_CMP__HELP

    o  REPORT_SHORTCUTS

    o  SERIALIZERS

    o  STR2REPORTSECTION_KEYS

    o  TIMEITNUMBER

    o  TMPFILENAME

    o  UNITS

    o  VERBOSITY_MINIMAL
    o  VERBOSITY_NORMAL
    o  VERBOSITY_DETAILS
    o  VERBOSITY_DEBUG



The following variables depend on previously defined variables,
hence their location at the end of this list:

    o  GRAPHS_FILENAME
    o  GRAPHS_DESCRIPTION
"""
import re


# command line arguments
# will be set to argparse.ArgumentParser(...).parse_args()
ARGS = None

# (pimydoc)cwc modules names
# ⋅
# ⋅cwc modules names start with the "wisteria.cwc" string (cf is_a_cwc_name())
# ⋅
# ⋅moduleininame        : e.g. "wisteria.cwc.pgnreader.cwc_default.chessgames"
# ⋅modulefullrealname   : e.g. "wisteria.cwc.pgnreader.cwc_default.ChessGames"
# ⋅waemodulename        : e.g. "wisteria.cwc.pgnreader.works_as_expected"
# ⋅classname            : e.g. "ChessGames" (NOT "chessgames")
# ⋅modulerealname       : e.g. "wisteria.cwc.pgnreader.cwc_default"
# ⋅
# ⋅- `moduleininame` are defined in config file;
# ⋅- conversion from `moduleininame` to `modulefullrealname` is defined in
# ⋅  data.py:DATA and is made by function
# ⋅  cwc_utils.py:moduleininame_to_modulefullrealname()
# ⋅- conversion from `modulefullrealname` to `modulerealname` is made by
# ⋅  function cwc_utils.py:modulefullrealname_to_modulerealname()
# ⋅- DATA keys (for cwc modules) use `moduleininame`, NOT `modulefullrealname`
CWC_MODULES = (
    ("wisteria.cwc.pgnreader.cwc_default.chessgames",
     "wisteria.cwc.pgnreader.cwc_default.ChessGames"),
    ("wisteria.cwc.pgnreader.cwc_iaswn.chessgames",
     "wisteria.cwc.pgnreader.cwc_iaswn.ChessGames"),

    ("wisteria.cwc.simple.cwc_default.simpleclass",
     "wisteria.cwc.simple.cwc_default.SimpleClass"),
    ("wisteria.cwc.simple.cwc_iaswn.simpleclass",
     "wisteria.cwc.simple.cwc_iaswn.SimpleClass"),
)

# dict storing all data used by the program.
#
# (pimydoc)DATA format
# ⋅Initialized by data.py::init_data()
# ⋅
# ⋅- for Python basic types, DATA values are the real value:
# ⋅    e.g. DATA["bool/false"] = False
# ⋅- for cwc modules, DATA keys are the ini name (not the real name)
# ⋅  and DATA values are the real name:
# ⋅    e.g. DATA["wisteria.cwc.pgnreader.iaswn.ChessGames"] =
# ⋅        "wisteria.cwc.pgnreader.iaswn.chessgames"
# ⋅- for third party types, DATA values are the real value:
# ⋅    e.g. DATA["dateutil(parser.parse)"] = dateutil.parser.parse("2021-03-04")
DATA = {}
# dict storing all data that the program can't use.
#
# * format: UNAVAILABLE_DATA[(str)data name] = string explaining why it's unavailable
# * initialized by data.py::init_data()
UNAVAILABLE_DATA = {}

# maximal console width used to display some debug messages:
DEBUG_CONSOLEWIDTH = 70

# default name for the config file.
DEFAULT_CONFIGFILE_NAME = "wisteria.ini"
# url of the default config file:
DEFAULT_CONFIGFILE_URL = "https://raw.githubusercontent.com/suizokukan/wisteria/main/wisteria.ini"

# default filename for --exportreport
DEFAULT_EXPORTREPORT_FILENAME = "report.md"

DEFAULT_REPORTFILE_NAME = "report.txt"

# value: rich.console.Console(file=...)
FILECONSOLE = None
# Both variables are initialized by main.py()
FILECONSOLE_FILEOBJECT = None

# imported serializers modules
MODULES = {}

# value of the --output argument
#  initialized by parse_output_argument(--output string)
# format:
#        ((bool)success, True if the --output string had been successfully parsed
#         (bool)output to the console ?,
#         (bool)output to the reportfile ?,
#         (str)reportfile open mode = 'a' or 'w',
#         (str)reportfile name,
#        )
OUTPUT = []

# (pimydoc)PLANNED_TRANSCODINGS
# ⋅list of str:
# ⋅    (str)serializer, (str)data_name, (str)fingerprint
# ⋅
# ⋅Initialized by results.py:init_planned_transcodings()
PLANNED_TRANSCODINGS = []

# = platform.system()
PLATFORM_SYSTEM = None

# progress bar length (in characters)
# 'None' is an accepted value and extends the length of the bar
# to the entire width of the terminal.
PROGRESSBAR_LENGTH = None

# (pimydoc)--cmp format
# ⋅
# ⋅(I) serializers
# ⋅Test one serializer alone(1) or one serializer against another serializer(2) or
# ⋅a serializer against all serializers(3) or all serializers(4) together.
# ⋅
# ⋅    (1) --cmp="jsonpickle(cwc)"
# ⋅    (2) --cmp="jsonpickle vs pickle (cwc)"
# ⋅    (3) --cmp="jsonpickle vs all (cwc)"
# ⋅    (4) --cmp="all vs all (cwc)"
# ⋅
# ⋅(II) data types:
# ⋅Instead of 'cwc' (=compare what's comparable)(a) you may want to test all data types
# ⋅but cwc(b) or data types defined in the config file(c) or absolutely all data types(d).
# ⋅
# ⋅    (a) --cmp="jsonpickle vs pickle (cwc)"
# ⋅    (b) --cmp="jsonpickle vs pickle (allbutcwc)"
# ⋅    (c) --cmp="jsonpickle vs pickle (ini)"
# ⋅    (d) --cmp="jsonpickle vs pickle (all)"
# ⋅
# ⋅NB: You may use 'vs' as well as 'against', as if:
# ⋅    --cmp="jsonpickle vs pickle (cwc)"
# ⋅NB: globs.py::REGEX_CMP defines exactly the expected format
# ⋅    globs.py::REGEX_CMP__HELP gives an idea of what is expected; this
# ⋅                              string is used as help message by the
# ⋅                              command line --help argument.
# ⋅
# regex used to parse the --cmp argument string
REGEX_CMP = re.compile(r"^\s*(?P<serializer1>[^\s\(\)]+)"
                       r"((\svs\s|\sversus\s|\sagainst\s)(?P<serializer2>[^\s\(\)]+))?"
                       r"(\s*\((?P<cmpdata>all|cwc|allbutcwc|ini)\))?\s*$")
# string used by --help
REGEX_CMP__HELP = "all|serializer1[vs all|serializer2][(cwc|allbutcwc|ini|all)]"

# REPORT_SHORTCUTS has two goals:
# (1) it translates a simple, human-readable string into a list of section parts
# (2) it gives all the accepted keywords that --report may understand.
REPORT_SHORTCUTS = {
    "bestof": "B1c;C2b;C2c;graphs;",
    "full": "titles;A;B;C;D1a;graphs;",
    "full+": "titles;A;B;C;D1b;graphs;",
    "full_debug": "titles;A;B;C;D;graphs;",
    "glance": "titles;B1b;C1a;C1b;C2a;C2b;C2c;graphs;",
    "laconic": "C;graphs;",
    "minimal": "C2c;graphs;",
    "none": "",
    }

# dict storing all serializers used by the program.
#
# * format: SERIALIZERS[(str)serializer name] = SerializerData object
# * initialized by serializers.py::init_serializers()
SERIALIZERS = {}
# dict storing all serializers that the program can't use.
#
# * format: UNAVAILABLE_SERIALIZERS[(str)serializer name] = SerializerData object
# * initialized by serializers.py::init_serializers()
UNAVAILABLE_SERIALIZERS = {}

#  globs.py:STR2REPORTSECTION_KEYS should be nothing but STR2REPORTSECTION.keys()
#
# Why those two variables ?
# Because the program needs to know STR2REPORTSECTION.keys() at step A
# before STR2REPORTSECTION has been initialized. Since this initialization
# requires modules that can't fit step A, we have to create a distinct list
# of keys.
#
# check_str2reportsection_keys() checks that keys from STR2REPORTSECTION and
# from STR2REPORTSECTION_KEYS are exactly the same.
STR2REPORTSECTION_KEYS = (
    'titles',
    'graphs',
    'A',
    'A1',
    'A2',
    'A3',
    'A4',
    'A5',
    'B',
    'B1',
    'B1a',
    'B1b',
    'B1c',
    'B1d',
    'B2',
    'B2a',
    'B2b',
    'C',
    'C1',
    'C1a',
    'C1b',
    'C2',
    'C2a',
    'C2b',
    'C2c',
    'D',
    'D1',
    'D1a',
    'D1b',
)

# number of times each serializer is called
TIMEITNUMBER = 1

# temp file default name
TMPFILENAME = "wisteria.tmp"

# units used in this project
UNITS = {'time': 'seconds',
         'string length': 'characters',
         'memory': 'bytes',
         }

# values defined for --verbosity:
VERBOSITY_MINIMAL = 0
VERBOSITY_NORMAL = 1
VERBOSITY_DETAILS = 2
VERBOSITY_DEBUG = 3


# =============================================================================
#
# The following variables depend on previously defined variables,
# hence their location at the end of this list.
#
# =============================================================================

# name of the graphs file
GRAPHS_FILENAME = "report__SUFFIX__.png"

# (pimydoc)GRAPHS_DESCRIPTION format
# ⋅Use GRAPHS_DESCRIPTION to store the description of each graph created by the
# ⋅report; each description is passed to hbar2png(). Note that
# ⋅len(GRAPHS_DESCRIPTION) gives the number of graphs to be created.
# ⋅
# ⋅- (str)attribute   : hbar2png will read results.hall[attribute]
# ⋅- (str)fmtstring   : format string to be applied to each value when printed
# ⋅                     on the graph; e.g. '{0}' or '{0:.1f}'
# ⋅- (int)value_coeff : each value will be multiplied by this number
# ⋅- (str)unit        : x unit
# ⋅- (str)title       : graph title
# ⋅- (str)filename    : file name to be written
GRAPHS_DESCRIPTION = (('encoding_time', "{0:.3f}", 1, UNITS['time'], 'Slowness',
                      GRAPHS_FILENAME.replace("__SUFFIX__", "1")),
                      ('mem_usage', "{0}", 1, UNITS['memory'], 'Memory Usage',
                       GRAPHS_FILENAME.replace("__SUFFIX__", "2")),
                      ('encoding_strlen', "{0}", 1, UNITS['string length'], 'Encoded String Length',
                       GRAPHS_FILENAME.replace("__SUFFIX__", "3")),
                      ('reversibility', "{0:.1f}", 100, "%", 'Coverage data (Reversibility)',
                       GRAPHS_FILENAME.replace("__SUFFIX__", "4")),)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
################################################################################
#    Wisteria Copyright (C) 2021 suizokukan
#    Contact: suizokukan _A.T._ orange dot fr
#
#    This file is part of Wisteria.
#    Wisteria is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Wisteria is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Wisteria.  If not, see <http://www.gnu.org/licenses/>.
################################################################################
"""
    Wisteria project : wisteria/cwc/cwc_utils.py

    Various functions to be used with cwc modules.


    (pimydoc)cwc modules names
    ⋅
    ⋅cwc modules names start with the "wisteria.cwc" string (cf is_a_cwc_name())
    ⋅
    ⋅moduleininame        : e.g. "wisteria.cwc.pgnreader.cwc_default.chessgames"
    ⋅modulefullrealname   : e.g. "wisteria.cwc.pgnreader.cwc_default.ChessGames"
    ⋅waemodulename        : e.g. "wisteria.cwc.pgnreader.works_as_expected"
    ⋅classname            : e.g. "ChessGames" (NOT "chessgames")
    ⋅modulerealname       : e.g. "wisteria.cwc.pgnreader.cwc_default"
    ⋅
    ⋅- `moduleininame` are defined in config file;
    ⋅- conversion from `moduleininame` to `modulefullrealname` is defined in
    ⋅  data.py:DATA and is made by function
    ⋅  cwc_utils.py:moduleininame_to_modulefullrealname()
    ⋅- conversion from `modulefullrealname` to `modulerealname` is made by
    ⋅  function cwc_utils.py:modulefullrealname_to_modulerealname()
    ⋅- DATA keys (for cwc modules) use `moduleininame`, NOT `modulefullrealname`


    ___________________________________________________________________________

    o  are_two_cwc_variants_of_the_same_cwc(name1, name2)
    o  count_dataobjs_number_without_cwc_variant(dataobjs)
    o  is_a_cwc_name(data_object_name)
    o  is_this_an_appropriate_module_for_serializer(data_name__module, serializer)
    o  modulefullrealname_to_classname(modulefullrealname)
    o  modulefullrealname_to_modulerealname(modulefullrealname)
    o  modulefullrealname_to_waemodulename(modulefullrealname)
    o  moduleininame_to_modulefullrealname(moduleininame)
    o  select__works_as_expected__function(data_object_name)
    o  serializer_is_compatible_with_dataobj(serializer, dataobj)
    o  shorten_cwc_name(cwc_name)
"""
import wisteria.globs


def are_two_cwc_variants_of_the_same_cwc(name1,
                                         name2):
    """
        are_two_cwc_variants_of_the_same_cwc()

        Return True if <name1> and <name2> are two cwc names and if the third element
        (=name.split(".")[2]) is the same, as in:
                'wisteria.cwc.pgnreader.cwc_default.ChessGames' and
                'wisteria.cwc.pgnreader.iaswn.ChessGames'
                ... where name.split(".")[2] is 'pgnreader'
    """
    if name1 == name2:
        return False
    if not is_a_cwc_name(name1):
        return False
    if not is_a_cwc_name(name2):
        return False

    return name1.split(".")[2] == name2.split(".")[2]


def count_dataobjs_number_without_cwc_variant(dataobjs):
    """
        count_dataobjs_number_without_cwc_variant()

        Return the number of data objects in <dataobjs> that are not a cwc variant
        of another data object in <dataobjs>.

        Unittest: CWCUtils.test_count_dataobjs_number_without_cwc_variant().

        if <dataobjs> is something like:
            {"wisteria.cwc.pgnreader.cwc_default.ChessGames": None,
             "wisteria.cwc.pgnreader.iaswn.ChessGames": None,
             "wisteria.cwc.pgnreader.xyz.ChessGames": None,
             "int": None,}      > the method will return 2

        _______________________________________________________________________

        ARGUMENT: (set of str)<dataobjs>, e.g.
            {'wisteria.cwc.pgnreader.cwc_default.ChessGames',
             'wisteria.cwc.pgnreader.iaswn.ChessGames'}

        RETURNED VALUE: (int)the number of data objects in <dataobjs> that are not
                        a cwc variant of another data object in <dataobjs>
    """
    cwc_variants = []
    res = 0
    for dataobj in dataobjs:
        if not is_a_cwc_name(dataobj):
            res += 1
        else:
            ok = True
            for cwc_variant in cwc_variants:
                if are_two_cwc_variants_of_the_same_cwc(cwc_variant,
                                                        dataobj):
                    ok = False
                    break
            if ok:
                res += 1
                cwc_variants.append(dataobj)

    return res


def is_a_cwc_name(data_object_name):
    """
        is_a_cwc_name()

        Return True if <data_object_name> starts with "wisteria.cwc".


        (pimydoc)cwc modules names
        ⋅
        ⋅cwc modules names start with the "wisteria.cwc" string (cf is_a_cwc_name())
        ⋅
        ⋅moduleininame        : e.g. "wisteria.cwc.pgnreader.cwc_default.chessgames"
        ⋅modulefullrealname   : e.g. "wisteria.cwc.pgnreader.cwc_default.ChessGames"
        ⋅waemodulename        : e.g. "wisteria.cwc.pgnreader.works_as_expected"
        ⋅classname            : e.g. "ChessGames" (NOT "chessgames")
        ⋅modulerealname       : e.g. "wisteria.cwc.pgnreader.cwc_default"
        ⋅
        ⋅- `moduleininame` are defined in config file;
        ⋅- conversion from `moduleininame` to `modulefullrealname` is defined in
        ⋅  data.py:DATA and is made by function
        ⋅  cwc_utils.py:moduleininame_to_modulefullrealname()
        ⋅- conversion from `modulefullrealname` to `modulerealname` is made by
        ⋅  function cwc_utils.py:modulefullrealname_to_modulerealname()
        ⋅- DATA keys (for cwc modules) use `moduleininame`, NOT `modulefullrealname`


        _______________________________________________________________________

        ARGUMENT: (str)data_object_name, the string to be checked.

        RETURNED VALUE: (bool)True if <data_object_name> is a cwc string.
    """
    return data_object_name.startswith("wisteria.cwc")


def is_this_an_appropriate_module_for_serializer(data_name__module,
                                                 serializer):
    """
        is_this_an_appropriate_module_for_serializer()

        Return True if <data_name__module> is a module appropriate to <serializer>:
            e.g. if data_name__module is "cwc.pgnreader.cwc_default" and
                 if SERIALIZERS[serializer].cwc is "default" > True

            e.g. if data_name__module is "cwc.pgnreader.cwc_default" and
                 if SERIALIZERS[serializer].cwc is "iaswn" > False - we skip.


        (pimydoc)cwc modules names
        ⋅
        ⋅cwc modules names start with the "wisteria.cwc" string (cf is_a_cwc_name())
        ⋅
        ⋅moduleininame        : e.g. "wisteria.cwc.pgnreader.cwc_default.chessgames"
        ⋅modulefullrealname   : e.g. "wisteria.cwc.pgnreader.cwc_default.ChessGames"
        ⋅waemodulename        : e.g. "wisteria.cwc.pgnreader.works_as_expected"
        ⋅classname            : e.g. "ChessGames" (NOT "chessgames")
        ⋅modulerealname       : e.g. "wisteria.cwc.pgnreader.cwc_default"
        ⋅
        ⋅- `moduleininame` are defined in config file;
        ⋅- conversion from `moduleininame` to `modulefullrealname` is defined in
        ⋅  data.py:DATA and is made by function
        ⋅  cwc_utils.py:moduleininame_to_modulefullrealname()
        ⋅- conversion from `modulefullrealname` to `modulerealname` is made by
        ⋅  function cwc_utils.py:modulefullrealname_to_modulerealname()
        ⋅- DATA keys (for cwc modules) use `moduleininame`, NOT `modulefullrealname`


        _______________________________________________________________________

        ARGUMENTS:
        o  (str)data_name__module
        o  (str)serializer

        RETURNED VALUE: (bool)True if <data_name__module> if appropriate to
                        <serializer>
    """
    return data_name__module.endswith(wisteria.globs.SERIALIZERS[serializer].cwc)


def modulefullrealname_to_classname(modulefullrealname):
    """
        modulefullrealname_to_classname()

        Convert module full real name into class name, something like:
                "wisteria.cwc.pgnreader.cwc_default.ChessGames" > "ChessGames"


        (pimydoc)cwc modules names
        ⋅
        ⋅cwc modules names start with the "wisteria.cwc" string (cf is_a_cwc_name())
        ⋅
        ⋅moduleininame        : e.g. "wisteria.cwc.pgnreader.cwc_default.chessgames"
        ⋅modulefullrealname   : e.g. "wisteria.cwc.pgnreader.cwc_default.ChessGames"
        ⋅waemodulename        : e.g. "wisteria.cwc.pgnreader.works_as_expected"
        ⋅classname            : e.g. "ChessGames" (NOT "chessgames")
        ⋅modulerealname       : e.g. "wisteria.cwc.pgnreader.cwc_default"
        ⋅
        ⋅- `moduleininame` are defined in config file;
        ⋅- conversion from `moduleininame` to `modulefullrealname` is defined in
        ⋅  data.py:DATA and is made by function
        ⋅  cwc_utils.py:moduleininame_to_modulefullrealname()
        ⋅- conversion from `modulefullrealname` to `modulerealname` is made by
        ⋅  function cwc_utils.py:modulefullrealname_to_modulerealname()
        ⋅- DATA keys (for cwc modules) use `moduleininame`, NOT `modulefullrealname`


        _______________________________________________________________________

        ARGUMENT: (str)modulefullrealname, the module full real name to be converted

        RETURNED VALUE: (str)class name
    """
    return modulefullrealname[modulefullrealname.rfind(".")+1:]


def modulefullrealname_to_modulerealname(modulefullrealname):
    """
        modulefullrealname_to_modulerealname()

        Convert module full real name into module read name, something like:
            "wisteria.cwc.pgnreader.cwc_default.ChessGames" > "wisteria.cwc.pgnreader.cwc_default"

        (pimydoc)cwc modules names
        ⋅
        ⋅cwc modules names start with the "wisteria.cwc" string (cf is_a_cwc_name())
        ⋅
        ⋅moduleininame        : e.g. "wisteria.cwc.pgnreader.cwc_default.chessgames"
        ⋅modulefullrealname   : e.g. "wisteria.cwc.pgnreader.cwc_default.ChessGames"
        ⋅waemodulename        : e.g. "wisteria.cwc.pgnreader.works_as_expected"
        ⋅classname            : e.g. "ChessGames" (NOT "chessgames")
        ⋅modulerealname       : e.g. "wisteria.cwc.pgnreader.cwc_default"
        ⋅
        ⋅- `moduleininame` are defined in config file;
        ⋅- conversion from `moduleininame` to `modulefullrealname` is defined in
        ⋅  data.py:DATA and is made by function
        ⋅  cwc_utils.py:moduleininame_to_modulefullrealname()
        ⋅- conversion from `modulefullrealname` to `modulerealname` is made by
        ⋅  function cwc_utils.py:modulefullrealname_to_modulerealname()
        ⋅- DATA keys (for cwc modules) use `moduleininame`, NOT `modulefullrealname`


        _______________________________________________________________________

        ARGUMENT: (str)modulefullrealname, the module full real name to be converted

        RETURNED VALUE: (str)module real name
    """
    return modulefullrealname[:modulefullrealname.rfind(".")]


def modulefullrealname_to_waemodulename(modulefullrealname):
    """
        modulefullrealname_to_waemodulename()

        Convert module full real name into a wae('works as expected') module name, something like:
                "wisteria.cwc.pgnreader.cwc_default.ChessGames" >
                        "wisteria.cwc.pgnreader.works_as_expected"

        _______________________________________________________________________

        ARGUMENT: (str)modulefullrealname, the module full real name to be converted

        RETURNED VALUE: (str)wae('works as expected') module name

    """
    name = modulefullrealname[:modulefullrealname.rfind(".")]
    name = name[:name.rfind(".")]
    return name + ".works_as_expected"


def moduleininame_to_modulefullrealname(moduleininame):
    """
        moduleininame_to_modulefullrealname()

        Convert module ini name into module full real name, something like:
            "cwc.pgnreader.cwc_default.chessgames" > "cwc.pgnreader.cwc_default.ChessGames"


        (pimydoc)cwc modules names
        ⋅
        ⋅cwc modules names start with the "wisteria.cwc" string (cf is_a_cwc_name())
        ⋅
        ⋅moduleininame        : e.g. "wisteria.cwc.pgnreader.cwc_default.chessgames"
        ⋅modulefullrealname   : e.g. "wisteria.cwc.pgnreader.cwc_default.ChessGames"
        ⋅waemodulename        : e.g. "wisteria.cwc.pgnreader.works_as_expected"
        ⋅classname            : e.g. "ChessGames" (NOT "chessgames")
        ⋅modulerealname       : e.g. "wisteria.cwc.pgnreader.cwc_default"
        ⋅
        ⋅- `moduleininame` are defined in config file;
        ⋅- conversion from `moduleininame` to `modulefullrealname` is defined in
        ⋅  data.py:DATA and is made by function
        ⋅  cwc_utils.py:moduleininame_to_modulefullrealname()
        ⋅- conversion from `modulefullrealname` to `modulerealname` is made by
        ⋅  function cwc_utils.py:modulefullrealname_to_modulerealname()
        ⋅- DATA keys (for cwc modules) use `moduleininame`, NOT `modulefullrealname`


        _______________________________________________________________________

        ARGUMENT: (str)moduleininame, the module ini name to be converted

        RETURNED VALUE: (str)module full real name
    """
    return wisteria.globs.DATA[moduleininame]


def select__works_as_expected__function(data_object_name):
    """
        select__works_as_expected__function()

        Return a "works_as_expected" function apppropriate to <data_object_name>.


        _______________________________________________________________________

        ARGUMENT: (str)<data_object_name>

        RETURNED VALUE: None or a callable
    """
    if not is_a_cwc_name(data_object_name):
        # <data_object_name> is a simple type like "int" (and not a cwc type)
        return None if not wisteria.data.works_as_expected(data_name=data_object_name) \
            else wisteria.data.works_as_expected

    # <data_object_name> is a cwc name:
    #  e.g. "wisteria.cwc.pgnreader.cwc_default.ChessGames" >
    #               "wisteria.cwc.pgnreader.works_as_expected.works_as_expected
    return getattr(wisteria.globs.MODULES[modulefullrealname_to_waemodulename(data_object_name)],
                   "works_as_expected")


def serializer_is_compatible_with_dataobj(serializer,
                                          dataobj):
    """
        serializer_is_compatible_with_dataobj()

        Return True if <serializer> can handle <dataobj>, i.e.
        (a) if <dataobj> is not a cwc name (since all serializers known how to handle
            non-cwc dataobjs.
        (b) if <dataobj> is a cwc name and if <serializer> is compatible with this <dataobj>,
            e.g. pickle knows how to handle "wisteria.cwc.pgnreader.cwc_default.ChessGames",
                        ... since SERIALIZERS["pickle"].cwc is "default"
                 iaswn knows how to handle "wisteria.cwc.pgnreader.iaswn.ChessGames",
                        ... since SERIALIZERS["iaswn"].cwc is "iaswn"
    """
    if not is_a_cwc_name(dataobj):
        return True

    return is_this_an_appropriate_module_for_serializer(
        modulefullrealname_to_modulerealname(dataobj), serializer)


def shorten_cwc_name(cwc_name):
    """
        shorten_cwc_name()

        Shorten <cwc_name> by removing a useless prefix.

        _______________________________________________________________________

        ARGUMENT: (str)cwc_name, the string to be shortened

        RESULT: (str)the shortened string
    """
    return "cwc:" + cwc_name[len("wisteria.cwc."):]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
################################################################################
#    Wisteria Copyright (C) 2021 suizokukan
#    Contact: suizokukan _A.T._ orange dot fr
#
#    This file is part of Wisteria.
#    Wisteria is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Wisteria is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Wisteria.  If not, see <http://www.gnu.org/licenses/>.
################################################################################
"""
    Wisteria project : wisteria/data.py

    All objects used as data by serializers are stored in the DATA dict.


    ___________________________________________________________________________

    o  anyfunc()
    o  constant_factory()
    o  init_data()
"""
import array
import calendar
import collections
import datetime
import decimal
import hashlib
import io
import numbers
import re
import time

import wisteria.globs
from wisteria.globs import TMPFILENAME, CWC_MODULES
from wisteria.classesexamples.simpleclasses import MetaClass, RegularClass
from wisteria.classesexamples.simpleclasses import RegularClassInheritedDict
from wisteria.classesexamples.simpleclasses import RegularClassInheritedList
from wisteria.cwc.cwc_utils import modulefullrealname_to_modulerealname
from wisteria.cwc.cwc_utils import modulefullrealname_to_waemodulename
from wisteria.utils import trytoimport
from wisteria.wisteriaerror import WisteriaError


def anyfunc():
    """
        anyfunc()

        Fake function used by the DATA dict.
    """


def constant_factory(value):
    """
        constant_factory()

        Used to initialize defaultdict objects
    """
    return lambda: value


def init_data():
    """
        init_data()

        Initialize wisteria.globs.DATA.

        PLEASE SYNCHRONIZE THIS LIST OF DATA WITH THE LIST IN THE
        CONFIG FILE(S) AND THE LIST IN README.MD.

        (pimydoc)DATA format
        ⋅Initialized by data.py::init_data()
        ⋅
        ⋅- for Python basic types, DATA values are the real value:
        ⋅    e.g. DATA["bool/false"] = False
        ⋅- for cwc modules, DATA keys are the ini name (not the real name)
        ⋅  and DATA values are the real name:
        ⋅    e.g. DATA["wisteria.cwc.pgnreader.iaswn.ChessGames"] =
        ⋅        "wisteria.cwc.pgnreader.iaswn.chessgames"
        ⋅- for third party types, DATA values are the real value:
        ⋅    e.g. DATA["dateutil(parser.parse)"] = dateutil.parser.parse("2021-03-04")
    """
    # -------------------------------------------------------------------------
    # -------------------- 1/3 basic Python types -----------------------------
    # -------------------------------------------------------------------------
    wisteria.globs.DATA = {
        "array(b)":  array.array('b', (-1, 2)),
        "array(b/empty)": array.array('b'),
        "array(b_unsigned)": array.array('b', (1, 2)),
        "array(b_unsigned/empty)": array.array('B'),
        "array(u)": array.array('u', 'hello \u2641'),
        "array(u/empty)": array.array('u'),
        "array(h)": array.array('h', (-1, 2)),
        "array(h/empty)": array.array('h'),
        "array(h_unsigned)": array.array('H', (1, 2)),
        "array(h_unsigned/empty)": array.array('H'),
        "array(i)": array.array('i', (-1, 2)),
        "array(i/empty)": array.array('i'),
        "array(i_unsigned)": array.array('I', (1, 2)),
        "array(i_unsigned/empty)": array.array('I'),
        "array(l)": array.array('l', [-1, 2, 3, 4, 5]),
        "array(l/empty)": array.array('l'),
        "array(l_unsigned)": array.array('L', (1, 2)),
        "array(l_unsigned/empty)": array.array('L'),
        "array(q)": array.array('q', (-1, 2)),
        "array(q/empty)": array.array('q'),
        "array(q_unsigned)": array.array('Q', (1, 2)),
        "array(q_unsigned/empty)": array.array('Q'),
        "array(f)": array.array('f', (1.3, float('nan'))),
        "array(f/empty)": array.array('f'),
        "array(d)": array.array('d', [1.0, 2.0, 3.14]),
        "array(d/empty)": array.array('d'),

        "bool/false": False,
        "bool/true": True,

        "bytearray": bytearray(b"123"),
        "bytearray(empty)": bytearray(),

        "bytes": b"123",
        "bytes(empty)": b"",

        "calendar(calendar(3))": calendar.Calendar(3),

        "collections.chainmap(empty)": collections.ChainMap(),
        "collections.chainmap": collections.ChainMap({1: 2}, {2: 3}),
        "collections.counter(empty)": collections.Counter(),
        "collections.counter": collections.Counter((1, 2)),
        "collections.defaultdict(empty)": collections.defaultdict(),
        "collections.defaultdict(func)": collections.defaultdict(constant_factory('<missing>')),
        "collections.defaultdict(int)": collections.defaultdict(int, {"a": 1, "b": 10}),
        "collections.defaultdict(list)": collections.defaultdict(list, {1: [2, 3]}),
        "collections.defaultdict(none)": collections.defaultdict(None, {1: [2, 3]}),
        "collections.defaultdict(set)": collections.defaultdict(set, {1: set((2, 3))}),
        "collections.deque(empty)": collections.deque(),
        "collections.deque": collections.deque((1, 2)),
        "collections.ordereddict(empty)": collections.OrderedDict(),
        "collections.ordereddict": collections.OrderedDict({1: 2}),

        "complex": 1+2j,

        "datetime(datetime.datetime)": datetime.datetime(2001, 12, 1),
        "datetime(datetime.timedelta)":
        datetime.datetime(2001, 12, 1) - datetime.datetime(2000, 12, 1),

        "decimal(0.5)": decimal.Decimal(0.5),
        "decimal(1/7)": decimal.Decimal(1) / decimal.Decimal(7),
        "decimal(nan)": decimal.Decimal('NaN'),
        "decimal(-infinity)": decimal.Decimal("-Infinity"),
        "decimal(+infinity)": decimal.Decimal("+Infinity"),

        "dict(keys/bool)": {False: "False", True: "True"},
        "dict(keys/float)": {1.1: "value1.1", 2.2: "value2.2"},
        "dict(keys/int)": {0: "value0", 1: "value1", 2: "value2"},
        "dict(keys/str)": {"key1": "value1", "key2": "value2"},
        "dict(keys/str+subdicts)": {"key1": "value1", "key2": "value2", "key3": {"key4": "key4", }},

        # DATA["file descriptor"] contains a file descriptor which will be closed
        # by
        #   pylint: disable=consider-using-with
        "file descriptor": open(TMPFILENAME,
                                encoding="utf-8"),

        "float": 1.1,
        "float(nan)": float('nan'),

        "frozenset": frozenset(("1", "2",)),
        "frozenset(empty)": frozenset(),

        "function": anyfunc,
        "function(python)": print,

        "hashlib(hashlib.sha1)": hashlib.sha1(b"some string"),
        "hashlib(hashlib.sha224)": hashlib.sha224(b"some string"),
        "hashlib(hashlib.sha256)": hashlib.sha256(b"some string"),
        "hashlib(hashlib.sha384)": hashlib.sha384(b"some string"),
        "hashlib(hashlib.sha512)": hashlib.sha512(b"some string"),

        "imported module": re,
        "imported module(class)": re.Pattern,
        "imported module(function)": re.sub,

        "int": 1,

        "io.string": io.StringIO(),
        "io.string(empty)": io.StringIO().write("string"),

        "list": ["1", "2", ],
        "list(empty)": [],
        "list(+sublists)": ["1", "2", ["3", ["4", ]]],

        "metaclass": MetaClass(),

        "memoryview": memoryview(b"123"),

        "none": None,

        "notimplemented": NotImplemented,

        "numbers(complex)": numbers.Complex,
        "numbers(integral)": numbers.Integral,
        "numbers(numbers)": numbers.Number(),
        "numbers(real)": numbers.Real,

        "pythonexception typeerror": TypeError,

        "range": range(1000),
        "range(empty)": range(0),

        "regularclass": RegularClass(),
        "regularclass(async_method)": RegularClass.async_method,
        "regularclass(class_method)": RegularClass.class_method,
        "regularclass(generator)": RegularClass.generator,
        "regularclass(method)": RegularClass.method,
        "regularclass(static_method)": RegularClass.static_method,
        "regularclassinheriteddict": RegularClassInheritedDict(),
        "regularclassinheritedlist": RegularClassInheritedList(),

        "re.match": re.match(".*", "abc"),
        "re.match(+flags)": re.match(".*", "abc", re.M),

        "re.pattern(bytes)": re.compile(".*"),
        "re.pattern(str)": re.compile(b".*"),

        "set": set(("1", "2",)),
        "set(empty)": set(),

        "str": "abc",
        "str(empty)": "",
        "str(long)": "abhg12234"*10000,
        "str(non ascii characters)": "êł¹@"+chr(0x1234)+chr(0x12345),

        "time(time.time)": time.time(),

        "tuple": ("1", "2",),
        "tuple(empty)": (),
        "tuple(+subtuples)": ("1", "2", ("3", ("4",))),

        "type(str)": str,
        "type(type(str))": type(str),

        # (pimydoc)demonstration_dataobj_a5
        # ⋅Data object used to show the encoded strings created by the serializers,
        # ⋅see A5 report section.
        # ⋅
        # ⋅The choice of data is very small because all serializers must be able
        # ⋅to encode it. By example, no None object because of Amazon Ion Python.
        "demonstration_dataobj_a5": {"key1": "value1",
                                     "key2": ["1", 2, False, True, ],
                                     "key3": {"subkey1": "subvalue1", },
                                     "key4": [[], [[]]],
                                     },
    }

    # -------------------------------------------------------------------------
    # --------------------------- 2/3 cwc modules -----------------------------
    # -------------------------------------------------------------------------
    # (pimydoc)cwc modules names
    # ⋅
    # ⋅cwc modules names start with the "wisteria.cwc" string (cf is_a_cwc_name())
    # ⋅
    # ⋅moduleininame        : e.g. "wisteria.cwc.pgnreader.cwc_default.chessgames"
    # ⋅modulefullrealname   : e.g. "wisteria.cwc.pgnreader.cwc_default.ChessGames"
    # ⋅waemodulename        : e.g. "wisteria.cwc.pgnreader.works_as_expected"
    # ⋅classname            : e.g. "ChessGames" (NOT "chessgames")
    # ⋅modulerealname       : e.g. "wisteria.cwc.pgnreader.cwc_default"
    # ⋅
    # ⋅- `moduleininame` are defined in config file;
    # ⋅- conversion from `moduleininame` to `modulefullrealname` is defined in
    # ⋅  data.py:DATA and is made by function
    # ⋅  cwc_utils.py:moduleininame_to_modulefullrealname()
    # ⋅- conversion from `modulefullrealname` to `modulerealname` is made by
    # ⋅  function cwc_utils.py:modulefullrealname_to_modulerealname()
    # ⋅- DATA keys (for cwc modules) use `moduleininame`, NOT `modulefullrealname`
    for cwc_moduleininame, cwc_modulefullrealname in CWC_MODULES:
        # main module, like "wisteria.cwc.pgnreader.cwc_default.ChessGames"
        if not trytoimport(modulefullrealname_to_modulerealname(cwc_modulefullrealname)):
            raise WisteriaError(
                "(ERRORID047) "
                "Internal error: can't import cwc module "
                f"'{modulefullrealname_to_modulerealname(cwc_modulefullrealname)}' .")
        wisteria.globs.DATA[cwc_moduleininame] = cwc_modulefullrealname

        # wae (=works as expected module), like "wisteria.cwc.pgnreader.works_as_expected"
        if not trytoimport(modulefullrealname_to_waemodulename(cwc_modulefullrealname)):
            raise WisteriaError(
                "(ERRORID048) "
                "Internal error: can't import 'works_as_expected' cwc module "
                f"'{modulefullrealname_to_waemodulename(cwc_modulefullrealname)}' .")

    # -------------------------------------------------------------------------
    # --------------------------- 3/3 third party types -----------------------
    # -------------------------------------------------------------------------
    try:
        # The following 'import' statement has deliberately placed here and
        # not at the beginning of the file.
        #   pylint: disable=import-outside-toplevel
        #   pylint: disable=import-error
        import dateutil.parser
        wisteria.globs.DATA["dateutil(parser.parse)"] = dateutil.parser.parse("2021-03-04")
    except ImportError:
        wisteria.globs.UNAVAILABLE_DATA["dateutil(parser.parse)"] = "missing package: dateutil"


# I don't want to split this function into different parts:
#   pylint: disable=too-many-return-statements
def works_as_expected(data_name,
                      obj=None):
    """
        works_as_expected()

        works_as_expected() function for basic types defined DATA/UNAVAILABLE_DATA.


        _______________________________________________________________________

        (pimydoc)works_as_expected arguments and returned value
        ⋅All works_as_expected() functions are supposed to (1) say if <data_name> is in
        ⋅the scope of this function (2) and say if <obj> works as expected.
        ⋅
        ⋅ARGUMENTS:
        ⋅    o  data_name:   (str)data_name of the <obj>ect
        ⋅    o  obj:         (None or any object) object to be checked
        ⋅
        ⋅RETURNED VALUE:
        ⋅    (<obj> is None)     (bool)<data_name> is known
        ⋅    (<obj> is not None) <obj> works as expected.
    """
    if data_name == "list":
        if obj is None:
            # yes, this data_name is known:
            return True
        # <obj> is not None: does <obj> work as expected ?
        obj.clear()
        if len(obj) != 0:
            return False
        obj.append("1")
        obj.append("2")
        if obj[1] != "2":
            return False
        return True

    # there's no
    #       if data_name == "collections.defaultdict(func)"
    # since no serializer can transcode these objects; their validation is
    # by consequence useless.

    if data_name == "collections.defaultdict(int)":
        if obj is None:
            # yes, this data_name is known:
            return True
        # <obj> is not None: does <obj> work as expected ?
        obj.clear()
        try:
            obj[555] += 1
        except KeyError:
            return False
        return obj[555] == 1

    if data_name == "collections.defaultdict(list)":
        if obj is None:
            # yes, this data_name is known:
            return True
        # <obj> is not None: does <obj> work as expected ?
        obj.clear()
        try:
            obj[555].append("555")
        except KeyError:
            return False
        return obj[555] == ["555", ]

    if data_name == "collections.defaultdict(none)":
        if obj is None:
            # yes, this data_name is known:
            return True
        # <obj> is not None: does <obj> work as expected ?
        obj.clear()
        try:
            obj[555].append("555")
        except KeyError:
            return True
        return False

    if data_name == "collections.defaultdict(set)":
        if obj is None:
            # yes, this data_name is known:
            return True
        # <obj> is not None: does <obj> work as expected ?
        obj.clear()
        try:
            obj[555].add("555")
        except KeyError:
            return False
        return obj[555] == set(("555",))

    # unknown data_name:
    return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
################################################################################
#    Wisteria Copyright (C) 2021 suizokukan
#    Contact: suizokukan _A.T._ orange dot fr
#
#    This file is part of Wisteria.
#    Wisteria is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Wisteria is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Wisteria.  If not, see <http://www.gnu.org/licenses/>.
################################################################################
"""
    Wisteria project : wisteria/dmfile.py

    DMFile class (=disk/in-memory file).

    Use DMFile to open on disk and in-memory files. You may use DMFile objects
    with Python 'with' statement.


    ___________________________________________________________________________

    o  DMFile class
"""
from io import StringIO


class DMFile:
    """
        DMFile class

        Disk/in-memory file. Very simple class allowing to open in-memory and on disk
        files.

        _______________________________________________________________________

        o  __enter__(self)
        o  __exit__(self, exc_type, exc_val, exc_tb)
        o  __init__(self, name, mode="r")
        o  __iter__(self)
        o  close(self)
        o  read(self)
        o  write(self, string)
    """
    def __enter__(self):
        """
            DMFile.__enter__()
        """
        return self

    def __exit__(self,
                 exc_type,
                 exc_val,
                 exc_tb):
        """
            DMFile.exit()
        """
        self.close()

    def __init__(self,
                 name,
                 mode="r"):
        """
            DMFile.__init__()
        """
        self.in_memory = name[0] == ":" and name[-1] == ":"
        if not self.in_memory:
            self.name = name
            #   pylint: disable=consider-using-with
            self.obj = open(name, mode=mode, encoding="utf-8")
        else:
            self.name = name[1:-1]
            self.obj = StringIO()

    def __iter__(self):
        """
            DMFile.__iter__()
        """
        for line in self.read().split("\n"):
            yield line

    def close(self):
        """
            DMFile.close()
        """
        if self.obj:
            self.obj.close()

    def read(self):
        """
            DMFile.read()
        """
        if not self.in_memory:
            self.obj.seek(0)
            return self.obj.read()

        # on disk:
        self.obj.seek(0)
        return self.obj.read()

    def write(self,
              string):
        """
            DMFile.write()
        """
        self.obj.write(string)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
################################################################################
#    Wisteria Copyright (C) 2021 suizokukan
#    Contact: suizokukan _A.T._ orange dot fr
#
#    This file is part of Wisteria.
#    Wisteria is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Wisteria is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Wisteria.  If not, see <http://www.gnu.org/licenses/>.
################################################################################
"""
    Wisteria project : wisteria/matplotgraphs.py

    Draw graphs using matplotlib.

    Code largely inspired by
        https://matplotlib.org/stable/gallery/lines_bars_and_markers/gradient_bar.html

    ___________________________________________________________________________

    o  gradient_image(axes, extent, direction=0.5, cmap_range=(0, 1), **kwargs)
    o  gradient_bar(axes, pos_x, pos_y, height=0.4, left=0)
    o  hbar2png(_data, filename, unit, title, fmtstring, value_coeff)
"""
from matplotlib import pyplot
import numpy as np

import wisteria.globs


def gradient_image(axes,
                   extent,
                   direction=0.5,
                   cmap_range=(0, 1),
                   **kwargs):
    """
        gradient_image()

        Draw a gradient image based on a colormap.

            Code largely inspired by
                https://matplotlib.org/stable/gallery/lines_bars_and_markers/gradient_bar.html

        _______________________________________________________________________

        ARGUMENTS:
        o  (matplotlib.axes._subplots.AxesSubplot) axes: axes to draw on
        o  (tuple of 4 floats)                   extent: the extent of the image,
                                                         (xmin, xmax, ymin, ymax),
                                                         expressed in in Axes coordinates
        o  (float)                            direction: direction of the gradient, a
                                                         number in range 0 (=vertical)
                                                         to 1 (=horizontal)
        o  (float, float)                    cmap_range: fraction (cmin, cmax) of the colormap that
                                                         should be used for the gradient, where the
                                                         complete colormap is (0, 1)
        o                                      **kwargs: other parameters are passed on to
                                                         `.Axes.imshow()`, in particular:
                o (str or Colormap)                cmap: Colormap instance or registered colormap
                                                         name used to map scalar data to colors.
    """
    phi = direction * np.pi / 2
    vector = np.array([np.cos(phi), np.sin(phi)])
    image = np.array([[vector @ [1, 0], vector @ [1, 1]],
                      [vector @ [0, 0], vector @ [0, 1]]])
    cmin, cmax = cmap_range
    image = cmin + (cmax - cmin) / image.max() * image
    # https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.imshow.html
    axes.imshow(image, extent=extent, interpolation='bicubic',
                vmin=0, vmax=1, **kwargs)


def gradient_bar(axes,
                 pos_x,
                 pos_y,
                 height=0.4,
                 left=0):
    """
        Draw a rectangle using gradient_image().

            Code largely inspired by
                https://matplotlib.org/stable/gallery/lines_bars_and_markers/gradient_bar.html

        _______________________________________________________________________

        ARGUMENTS:
        o  (tuple of float-s)pos_x : for each item, rightest x position (=right)
        o  (tuple of float-s)pos_y : for each item, highest y position (=bottom; here top > bottom)
        o  (float)height: height of the bar
        o  (float)left  : leftest x position
    """
    for right, bottom in zip(pos_x, pos_y):
        top = bottom + height

        # Pylint doesn't seem to know pyplot.cm.* members:
        #   pylint: disable=no-member
        gradient_image(axes, extent=(left, right, bottom - 0.4, top),
                       cmap=pyplot.cm.Reds_r, cmap_range=(0, 0.8))


def hbar2png(results_hall_attribute,
             filename,
             unit,
             title,
             fmtstring,
             value_coeff):
    """
            Create a graph with horizontal bars from <results_hall_attribute> and write it in
            <filename>.

            Code largely inspired by
                https://matplotlib.org/stable/gallery/lines_bars_and_markers/gradient_bar.html

            ___________________________________________________________________

            ARGUMENTS:
            o  (list of (value, serializer name))results_hall_attribute: content of
                                                                         results.hall[attribute]
            o  (str)                                           filename: path to the file to be
                                                                         written
            o  (str)                                               unit: unit of values to be
                                                                         printed on the graph
            o  (str)                                              title: title of the graph
            o  (str)                                          fmtstring: format string to be applied
                                                                         to each value when printed
                                                                         on the graph; e.g. '{0}' or
                                                                         '{0:.1f}' .
            o  (float)                                      value_coeff: each value will be
                                                                         multiplied by this number.
    """
    pyplot.rcdefaults()
    pyplot.rcParams.update({'axes.labelsize': 'large',
                            'xtick.labelsize': 'x-small',
                            'ytick.labelsize': 'large',
                            })

    length = len(results_hall_attribute)
    values = tuple(item[0]*value_coeff for item in results_hall_attribute)
    serializers_names = tuple(
        wisteria.globs.SERIALIZERS[item[1]].human_name for item in results_hall_attribute)

    if len(tuple(0 for value in values if value < 0.001)) == len(values):
        # A special case: all values are (nearly or exactly) equal to 0,
        # hence we can't apply the usual rule to set xmax:
        xlim = 0, 1  # = xmin, xmax
    else:
        xlim = 0, max(values)*1.2  # = xmin, xmax
    ylim = -1, length  # = ymin, ymax

    fig, axes = pyplot.subplots()
    fig.subplots_adjust(left=0.35)
    axes.set(xlim=xlim, ylim=ylim, autoscale_on=False)

    # background image:
    # Pylint doesn't seem to know pyplot.cm.* members:
    #   pylint: disable=no-member
    gradient_image(axes, direction=0.1, extent=(0, 1, 0, 1), transform=axes.transAxes,
                   cmap=pyplot.cm.Blues, cmap_range=(0.1, 0.6))

    # bars:
    gradient_bar(axes, values, range(length))
    axes.set_title(title)
    for value_index, value in enumerate(values):
        axes.text(value,
                  value_index,
                  "  " + fmtstring.format(value),
                  color='red',
                  va='center',
                  fontsize=9, fontweight='normal')

    axes.set_xlabel(unit)
    axes.set_aspect('auto')
    pyplot.yticks(range(length), serializers_names)
    axes.set_yticks(range(length))

    pyplot.savefig(filename)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
################################################################################
#    Wisteria Copyright (C) 2021 suizokukan
#    Contact: suizokukan _A.T._ orange dot fr
#
#    This file is part of Wisteria.
#    Wisteria is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Wisteria is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Wisteria.  If not, see <http://www.gnu.org/licenses/>.
################################################################################
"""
    Wisteria project : wisteria/cwc/pgnreader/works_as_expected.py

    initialize() and works_as_expected() functions for all cwc/pgnreader/ classes.


    ___________________________________________________________________________

    o  initialize(obj)
    o  works_as_expected(obj)
"""
from wisteria.dmfile import DMFile


def initialize(obj):
    """
        initialize() function

        _______________________________________________________________________

        ARGUMENT: <obj>, the object to be initialized

        RETURNED OBJECT: <obj>, the now initialized object
    """
    data = """
[Event "Aimchess US Rapid Prelim"]
[Site "chess24.com INT"]
[Date "2021.08.29"]
[Round "6.6"]
[White "Dominguez Perez, Leinier"]
[Black "Aronian, Levon"]
[Result "0-1"]
[WhiteTitle "GM"]
[BlackTitle "GM"]
[WhiteElo "2758"]
[BlackElo "2782"]
[ECO "C53"]
[Opening "Giuoco Piano"]
[WhiteFideId "3503240"]
[BlackFideId "13300474"]
[EventDate "2021.08.28"]

1. e4 e5 2. Nf3 Nc6 3. Bc4 Bc5 4. c3 Nf6 5. d3 O-O 6. O-O d5 7. exd5 Nxd5 8. Re1
Bg4 9. Nbd2 Nb6 10. h3 Bh5 11. Bb3 Qxd3 12. Nxe5 Qf5 13. Nef3 Rad8 14. Qe2 Nd5
15. Ne4 Bxf3 16. Qxf3 Qxf3 17. gxf3 Bb6 18. a4 h6 19. Rd1 Nde7 20. Rxd8 Rxd8 21.
Kf1 Kf8 22. Be3 Nf5 23. Bxb6 axb6 24. Rd1 Rxd1+ 25. Bxd1 Ne5 26. Ng3 Nh4 27. Ke2
Nhxf3 28. Ke3 Ng1 29. b3 Nxh3 30. f4 Ng6 31. Nh5 Ne7 32. Bg4 Nd5+ 33. Kf3 g6 34.
Bxh3 gxh5 35. Bc8 Nxc3 36. Bxb7 Ke7 37. Ke3 Kd6 38. Kd4 Na2 39. Ke4 Nb4 40. Kf5
Nd5 41. Ba6 h4 42. Kg4 h3 43. Be2 h2 44. Bf3 Kc5 45. f5 Nf6+ 46. Kf4 Kb4 47. Ke5
Ng4+ 48. Kd5 Kxb3 0-1

[Event "Vienna"]
[Site "Vienna AUH"]
[Date "1882.05.17"]
[EventDate "1882.05.10"]
[Round "7"]
[Result "1-0"]
[White "Wilhelm Steinitz"]
[Black "Bernhard Fleissig"]
[ECO "C00"]
[WhiteElo "?"]
[BlackElo "?"]
[PlyCount "77"]

1. e4 e6 2. e5 d5 3. exd6 Bxd6 4. d4 Ne7 5. Bd3 Ng6 6. Nf3 Nc6
7. Nc3 Nb4 8. Bc4 c6 9. Ne4 Bc7 10. O-O O-O 11. Re1 Nd5
12. Nc5 Nh4 13. Ne5 Nf5 14. c3 Bxe5 15. Rxe5 Nf6 16. Re1 h6
17. Qf3 Nd5 18. Bb3 b6 19. Nd3 Ba6 20. Ne5 Rc8 21. Bc2 Nfe7
22. Qg3 Kh8 23. Qh4 Kg8 24. Qg3 Kh8 25. Qh3 Ng8 26. Qh5 Rc7
27. Bd2 Ndf6 28. Qh3 Nd5 29. c4 Ndf6 30. Rad1 Qe8 31. Bf4 Rc8
32. Qa3 Bb7 33. Qxa7 Ba8 34. Qxb6 g5 35. Bg3 Nd7 36. Qb3 f5
37. f3 Kg7 38. c5 Ndf6 39. Nc4 1-0
    """
    with DMFile(":memory:") as src:
        src.write(data)
        if obj.read_pgn(src) is False:
            return None
    return obj


def works_as_expected(data_name,
                      obj=None):
    """
        works_as_expected()

        works_as_expected() function for basic types defined DATA/UNAVAILABLE_DATA.


        _______________________________________________________________________

        (pimydoc)works_as_expected arguments and returned value
        ⋅All works_as_expected() functions are supposed to (1) say if <data_name> is in
        ⋅the scope of this function (2) and say if <obj> works as expected.
        ⋅
        ⋅ARGUMENTS:
        ⋅    o  data_name:   (str)data_name of the <obj>ect
        ⋅    o  obj:         (None or any object) object to be checked
        ⋅
        ⋅RETURNED VALUE:
        ⋅    (<obj> is None)     (bool)<data_name> is known
        ⋅    (<obj> is not None) <obj> works as expected.
    """
    if obj is None:
        # for CWC objects, this case is never reached.
        return True

    if len(obj) != 2:
        return False
    return obj[1].board.human_repr() == \
        "♝_♜_♛♜♞_\n" \
        "______♚_\n" \
        "__♟_♟♞_♟\n" \
        "__♙__♟♟_\n" \
        "__♘♙____\n" \
        "_♕___♙♗_\n" \
        "♙♙♗___♙♙\n" \
        "___♖♖_♔_"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
################################################################################
#    Wisteria Copyright (C) 2021 suizokukan
#    Contact: suizokukan _A.T._ orange dot fr
#
#    This file is part of Wisteria.
#    Wisteria is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Wisteria is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Wisteria.  If not, see <http://www.gnu.org/licenses/>.
################################################################################
"""
    Wisteria project : wisteria/wisteria/cwc/pgnreader/s_iaswn.py

    cwc/PGN reader+writer specific to Iaswn serializer. The following types are used:
    - bool, int, str
    - dict, list, tuple
        Note that dicts may have non string keys.
    ___________________________________________________________________________

    PGN reader+writer (https://en.wikipedia.org/wiki/Portable_Game_Notation).
    Regular chess only; no sophisticated rules (but castling, en passant and
    promotion are known). This is a very simple version of what a real PGN reader/writer might
    look like: no optimized storage, no move validation.

    With this module you can read a .pgn file, play the game move by move and write the game in
    another file. PGN-files containing several games are correctly read and written.

    No move validation: a move is always assumed to be valid.
    Data other than the moves (events, player names, dates and so on) are
    stored but are not used.

    A PGN viewer allowing to show giant games (cf tests/game5.pgn) may be find at:
            https://fr.chesstempo.com/pgn-viewer/


    ___________________________________________________________________________

    o  COLOR_NOCOLOR, COLOR_WHITE, COLOR_BLACK
    o  invert_color(color)
    o  PIECENATURE_NOPIECE, PIECENATURE_PAWN, PIECENATURE_ROOK, PIECENATURE_KNIGHT
       PIECENATURE_BISHOP, PIECENATURE_QUEEN, PIECENATURE_KING
    o  PIECENATURE2ALGEBRICNOTATION / ALGEBRICNOTATION2PIECENATURE
    o  MOVETYPE_SINGLE, MOVETYPE_CAPTURE, MOVETYPE_CASTLING

    o  ChessError class
    o  ChessGameTags class
    o  ChessPiece class
    o  ChessMove class
    o  ChessListOfMoves class
    o  ChessGameStatus class
    o  ChessBoard class
    o  ChessGame class
    o  ChessGames class

    o  fill(obj)
    o  works_as_expected()
"""
# For this demonstration file, some writing rules are not respected:
#   pylint: disable=invalid-name
#   pylint: disable=too-few-public-methods
#   pylint: disable=too-many-nested-blocks
#   pylint: disable=too-many-arguments
import copy
import re
from dataclasses import dataclass

from iaswn.iaswn import Iaswn

COLOR_NOCOLOR = 0
COLOR_WHITE = 1
COLOR_BLACK = 2


def invert_color(color):
    """invert_color((int)color)"""
    if color == COLOR_WHITE:
        return COLOR_BLACK
    return COLOR_WHITE


PIECENATURE_NOPIECE = 0
PIECENATURE_PAWN = 1
PIECENATURE_ROOK = 2
PIECENATURE_KNIGHT = 3
PIECENATURE_BISHOP = 4
PIECENATURE_QUEEN = 5
PIECENATURE_KING = 6
PIECENATURE2ALGEBRICNOTATION = {
    PIECENATURE_ROOK: 'R',
    PIECENATURE_KNIGHT: 'N',
    PIECENATURE_BISHOP: 'B',
    PIECENATURE_QUEEN: 'Q',
    PIECENATURE_KING: 'K'}
ALGEBRICNOTATION2PIECENATURE = {value: key for key, value in PIECENATURE2ALGEBRICNOTATION.items()}

MOVETYPE_SINGLE = 0
MOVETYPE_CAPTURE = 1
MOVETYPE_CASTLING_OO = 2
MOVETYPE_CASTLING_OOO = 3


class ChessError(Exception, Iaswn):
    """
        ChessError class

        Unique error raised by the program.
    """


class ChessGameTags(dict, Iaswn):
    """
        ChessGameTags class

        Use this class to store tags read in .pgn files like these ones:

            [Event "F/S Return Match"]
            [Site "Belgrade, Serbia JUG"]
            [Date "1992.11.04"]
            [Round "29"]
            [White "Fischer, Robert J."]
            [Black "Spassky, Boris V."]
            [Result "1/2-1/2"]


        _______________________________________________________________________

        o  __eq__(self, other)
        o  __init__(self, datadict=None)
        o  __repr__(self)
    """
    def __eq__(self,
               other):
        """
            ChessGamesTags.__eq__()

            Stricly speaking, not required to read/write PGN files but required
            to check that everything "works as expected" (see cwc validation).

            ___________________________________________________________________

            ARGUMENT: (ChessGamesTags)other, the object compared to self

            RETURNED VALUE: (bool)True if other == self
        """
        return dict.__eq__(self, other)

    def __init__(self,
                 datadict=None):
        """ChessGameTags.__init__()"""
        dict.__init__(self)
        Iaswn.__init__(self)
        if datadict:
            for key, value in datadict:
                self[key] = value

    def __repr__(self):
        """ChessGameTags.__repr__()"""
        return "; ".join(f"{key}: {value}" for key, value in self.items())


@dataclass
class ChessPiece(Iaswn):
    """
        ChessPiece class

        _______________________________________________________________________

        o  piece2unicode/unicode2piece

        o  (int)color, e.g. COLOR_WHITE
        o  (int)nature, e.g. PIECENATURE_PAWN

        o  __eq__(self, other)
    """
    color: int = COLOR_NOCOLOR
    nature: int = PIECENATURE_NOPIECE

    piece2unicode = {
        (PIECENATURE_NOPIECE, COLOR_NOCOLOR): '_',
        (PIECENATURE_PAWN, COLOR_WHITE): '♙',
        (PIECENATURE_PAWN, COLOR_BLACK): '♟',
        (PIECENATURE_ROOK, COLOR_WHITE): '♖',
        (PIECENATURE_ROOK, COLOR_BLACK): '♜',
        (PIECENATURE_KNIGHT, COLOR_WHITE): '♘',
        (PIECENATURE_KNIGHT, COLOR_BLACK): '♞',
        (PIECENATURE_BISHOP, COLOR_WHITE): '♗',
        (PIECENATURE_BISHOP, COLOR_BLACK): '♝',
        (PIECENATURE_QUEEN, COLOR_WHITE): '♕',
        (PIECENATURE_QUEEN, COLOR_BLACK): '♛',
        (PIECENATURE_KING, COLOR_WHITE): '♔',
        (PIECENATURE_KING, COLOR_BLACK): '♚',
        }
    unicode2piece = {value: key for key, value in piece2unicode.items()}

    def __eq__(self,
               other):
        """ChessPiece.__eq__()"""
        return self.color == other.color and self.nature == other.nature

    def __repr__(self):
        """ChessPiece.__repr__()"""
        return f"ChessPiece: {self.color=}; {self.nature=}"

    def human_repr(self):
        """ChessPiece.human_repr()"""
        return ChessPiece.piece2unicode[self.nature, self.color]

    def init_from_unicode_string(self,
                                 string):
        """ChessPiece.init_from_unicode_string()"""
        self.nature, self.color = ChessPiece.unicode2piece[string]
        return self

    def is_empty(self):
        """ChessPiece.is_empty()"""
        return self.color == COLOR_NOCOLOR and self.nature == PIECENATURE_NOPIECE


@dataclass
class ChessMove(Iaswn):
    """
        ChessMove class


        _______________________________________________________________________

        o  (int)movetype, e.g. MOVETYPE_SINGLE
        o  (iterable)beforeafter_coord_piece1, e.g. ((0,0), (1,1))
        o  (iterable)beforeafter_coord_piece2, e.g. ((0,0), (1,1))
        o  (None|int)promotion, e.g. PIECENATURE_QUEEN
        o  (bool)enpassant
        o  (str)str_game_result, e.g. "0-1"

        o  __eq__(self, other)
        o  __repr__(self)
    """
    beforeafter_coord_piece1: tuple
    beforeafter_coord_piece2: tuple
    movetype: int = MOVETYPE_SINGLE
    promotion: int = None
    enpassant: bool = False
    str_game_result: str = None

    def __eq__(self,
               other):
        """
            ChessMove.__eq__()

            Stricly speaking, not required to read/write PGN files but required
            to check that everything "works as expected" (see cwc validation).

            ___________________________________________________________________

            ARGUMENT: (ChessMove)other, the object compared to self

            RETURNED VALUE: (bool)True if other == self
        """
        return self.movetype == other.movetype and \
            self.beforeafter_coord_piece1 == other.beforeafter_coord_piece1 and \
            self.beforeafter_coord_piece2 == other.beforeafter_coord_piece2 and \
            self.promotion == other.promotion and \
            self.enpassant == other.enpassant and \
            self.str_game_result == other.str_game_result

    def __repr__(self):
        """ChessMove.__repr__()"""
        return f"{self.movetype=}; " \
            f"{self.beforeafter_coord_piece1=}; {self.beforeafter_coord_piece2}; " \
            f"{self.promotion=}; {self.enpassant=}; {self.str_game_result=};"


@dataclass
class ChessListOfMoves(list, Iaswn):
    """
        ChessListOfMoves class

        A list of ((int)doublemove_number, (int)self.next_player, (ChessMove)move)


        _______________________________________________________________________

        o  (int)next_player
        o  (int)doublemove_number

        o  __eq__(self, other)
        o  __repr__(self)
        o  add_move(self, move)
        o  who_plays(self)
    """
    next_player: int = COLOR_WHITE
    doublemove_number: int = 1

    def __eq__(self,
               other):
        """
            ChessListOfMoves.__eq__()

            Stricly speaking, not required to read/write PGN files but required
            to check that everything "works as expected" (see cwc validation).

            ___________________________________________________________________

            ARGUMENT: (ChessListOfMoves)other, the object compared to self

            RETURNED VALUE: (bool)True if other == self
        """
        return list.__eq__(self, other) and \
            self.next_player == other.next_player and \
            self.doublemove_number == other.doublemove_number

    def __repr__(self):
        """ChessListOfMoves.__repr__()"""
        return f"list of moves: {self.next_player=}; {self.doublemove_number=}; " + \
            "; ".join(repr(move) for move in self)

    def add_move(self,
                 move):
        """
            ChessListOfMoves.add_move()

            This move is played by self.next_player
        """
        self.append((self.doublemove_number, self.next_player, move))

        if self.next_player == COLOR_WHITE:
            self.next_player = COLOR_BLACK
        else:
            self.next_player = COLOR_WHITE
            self.doublemove_number += 1

    def who_plays(self):
        """ChessListOfMoves.who_plays()"""
        return self.next_player


@dataclass
class ChessGameStatus(Iaswn):
    """
        ChessGameStatus class


        _______________________________________________________________________

        o  (bool)game_is_over
        o  (int)who_won, e.g. COLOR_BLACK

        o  __eq__(self, other)
        o  __repr__(self)
        o  copy(self)
        o  update_from_pgn_string(self,
                               status_string,
                               current_player)
    """
    game_is_over: bool = False
    who_won: int = COLOR_NOCOLOR

    def __eq__(self,
               other):
        """
            ChessGameStatus.__eq__()

            Stricly speaking, not required to read/write PGN files but required
            to check that everything "works as expected" (see cwc validation).

            ___________________________________________________________________

            ARGUMENT: (ChessGameStatus)other, the object compared to self

            RETURNED VALUE: (bool)True if other == self
        """
        return self.game_is_over == other.game_is_over and \
            self.who_won == other.who_won

    def __repr__(self):
        """ChessGameStatus.__repr__()"""
        return f"{self.game_is_over=}; {self.who_won=}"

    def copy(self):
        """ChessGameStatus.copy()"""
        return ChessGameStatus(game_is_over=self.game_is_over,
                               who_won=self.who_won)

    def update_from_pgn_string(self,
                               status_string,
                               current_player):
        """
            ChessGameStatus.update_from_pgn_string()

            Beware ! if you modify this method, please update
                    ChessGame.regex_pgn_listofmoves['game_result']
        """
        if status_string == "1/2-1/2":
            self.game_is_over = True
            self.who_won = COLOR_NOCOLOR
        elif status_string == "0-1":
            self.game_is_over = True
            self.who_won = COLOR_BLACK
        elif status_string == "1-0":
            self.game_is_over = True
            self.who_won = COLOR_WHITE
        elif status_string == "*":
            self.game_is_over = False
            self.who_won = None
        elif status_string == "++":
            self.game_is_over = True
            self.who_won = current_player
        else:
            raise ChessError(f"Unknown status string '{status_string}'.")


class ChessBoard(Iaswn):
    """
        ChessBoard class


        _______________________________________________________________________

        o  moves_descr

        o  (dict)board
        o  (ChessGameStatus)pieces_status

        o  __eq__(self, other)
        o  __init__(self)
        o  __repr__(self)
        o  copy(self)
        o  get_king_coord(self, color)
        o  get_xy(self)
        o  human_repr(self)
        o  init_from_unicode_string(self, string)
        o  is_empty_or_is_this_piece(self, xy, piece)
        o  is_kingpinned(self, xy0, xy1)
        o  iter_through_all_squares()
        o  set_startpos(self)
        o  set_xy(self, xy, value)
        o  set_xy_empty(self, xy)
        o  update_by_playing_a_move(self, move)
        o  which_piece_could_go_to(self, piece, coord_after, movetype)
        o  who_attacks(self, xy)
        o  xy_is_off_the_board(xy)
    """
    # moves description for all pieces but knight and pawn
    # moves[nature] = (deltamax, ((deltax, deltay), ...))
    moves_descr = {PIECENATURE_BISHOP: (8,
                                        ((+1, -1), (+1, +1), (-1, +1), (-1, -1)),),
                   PIECENATURE_ROOK: (8,
                                      ((0, -1), (0, +1), (-1, 0), (+1, 0)),),
                   PIECENATURE_QUEEN: (8,
                                       ((+1, -1), (+1, +1), (-1, +1), (-1, -1),
                                        (0, -1), (0, +1), (-1, 0), (+1, 0)),),
                   PIECENATURE_KING: (2,
                                      ((+1, -1), (+1, +1), (-1, +1), (-1, -1),
                                       (0, -1), (0, +1), (-1, 0), (+1, 0)),),
                   }

    def __eq__(self,
               other):
        """
            ChessBoard.__eq__()

            Stricly speaking, not required to read/write PGN files but required
            to check that everything "works as expected" (see cwc validation).

            ___________________________________________________________________

            ARGUMENT: (ChessBoard)other, the object compared to self

            RETURNED VALUE: (bool)True if other == self
        """
        return self.board == other.board and \
            self.pieces_status == other.pieces_status

    def __init__(self):
        """ChessBoard.__init__()"""
        Iaswn.__init__(self)
        self.board = {}  # cf .get_xy(), set_xy()
        self.pieces_status = ChessGameStatus()

        self.set_startpos()

    def __repr__(self):
        """ChessBoard.__repr__()"""
        return self.human_repr() + repr(self.pieces_status)

    def copy(self):
        """ChessBoard.copy()"""
        res = ChessBoard()
        res.board = copy.deepcopy(self.board)
        res.pieces_status = self.pieces_status.copy()
        return res

    def get_king_coord(self,
                       color):
        """
            ChessBoard.copy()

            Return the (x, y) of the <color>(white/black) king
        """
        for xy in self.iter_through_all_squares():
            obj = self.get_xy(xy)
            if obj.nature == PIECENATURE_KING and \
               obj.color == color:
                return xy
        raise ChessError(f"No king (color: {color}): {self.human_repr()}")

    def get_xy(self,
               xy):
        """ChessBoard.get_xy()"""
        return self.board[xy[0], xy[1]]

    def human_repr(self):
        """ChessBoard.human_repr()"""
        res = []
        for y in range(8):
            res.append("".join(self.get_xy((x, y)).human_repr() for x in range(8)))
        return "\n".join(res)

    def init_from_unicode_string(self,
                                 string):
        """ChessBoard.init_from_unicode_string()"""
        index = 0
        for xy in self.iter_through_all_squares():
            self.set_xy(xy, ChessPiece().init_from_unicode_string(string[index]))
            index += 1

    def is_empty_or_is_this_piece(self,
                                  xy,
                                  piece):
        """
            ChessBoard.get_xy()

            Return True if [x, y] is an empty square OR if [x, y] is <piece>
        """
        obj = self.get_xy(xy)
        return obj.is_empty() or obj == piece

    def is_kingpinned(self,
                      xy0,
                      xy1):
        """
            ChessBoard.is_kingpinned()

            Return True if board[x0, y0] moving to bord[x1, y1] is impossible since board[x0, y0]
            is a king pinned piece.
        """
        # <piece> is the piece that may be pinned:
        piece = self.get_xy(xy0)
        # <king> is the king of <piece>:
        king = self.get_king_coord(piece.color)
        len0 = len(self.who_attacks(king))

        # <board> is an alternative board with <piece> being moved to (x1, y1):
        _board = self.copy()
        _board.set_xy_empty(xy0)
        _board.set_xy(xy1, piece)
        # the king may have just moved, so we update his position:
        king = _board.get_king_coord(piece.color)
        len1 = len(_board.who_attacks(king))

        return len0 < len1

    @staticmethod
    def iter_through_all_squares():
        """ChessBoard.iter_through_all_squares()"""
        for y in range(8):
            for x in range(8):
                yield x, y

    def set_startpos(self):
        """ChessBoard.set_startpos()"""
        self.board = {}
        self.pieces_status = ChessGameStatus()

        for xy in self.iter_through_all_squares():
            self.set_xy(xy, ChessPiece())

        self.init_from_unicode_string("♜♞♝♛♚♝♞♜"
                                      "♟♟♟♟♟♟♟♟"
                                      "________"
                                      "________"
                                      "________"
                                      "________"
                                      "♙♙♙♙♙♙♙♙"
                                      "♖♘♗♕♔♗♘♖")

    def set_xy(self,
               xy,
               value):
        """ChessBoard.set_xy()"""
        self.board[xy[0], xy[1]] = value

    def set_xy_empty(self,
                     xy):
        """ChessBoard.set_xy_empty()"""
        self.set_xy(xy, ChessPiece())

    def update_by_playing_a_move(self,
                                 move):
        """
            ChessBoard.update_by_playing_a_move()

            Modify <self> by playing a <move>.
        """
        if move.movetype in (MOVETYPE_SINGLE, MOVETYPE_CAPTURE):
            before, after = move.beforeafter_coord_piece1
            piece = self.get_xy(before)
            self.set_xy_empty(before)
            if not move.promotion:
                self.set_xy(after, piece)
            else:
                self.set_xy(after, ChessPiece(nature=move.promotion,
                                              color=piece.color))

            if move.enpassant:
                if piece.color == COLOR_WHITE:
                    self.set_xy_empty((after[0], after[1]+1))
                else:
                    self.set_xy_empty((after[0], after[1]-1))

        elif move.movetype in (MOVETYPE_CASTLING_OO, MOVETYPE_CASTLING_OOO):
            # king:
            before, after = move.beforeafter_coord_piece1
            piece = self.get_xy(before)
            self.set_xy_empty(before)
            self.set_xy(after, piece)
            # rook:
            before, after = move.beforeafter_coord_piece2
            piece = self.get_xy(before)
            self.set_xy_empty(before)
            self.set_xy(after, piece)

    def which_piece_could_go_to(self,
                                piece,
                                coord_after,
                                movetype):
        """
            ChessBoard.which_piece_could_go_to()

            Where does the <piece> that will arrive on <coord_after> come from?

            More than one result may be returned !
        """
        def add_to_res_if_rightpiece_notpinned(_xy, coord_after, piece):
            if not ChessBoard.xy_is_off_the_board(_xy) and \
               self.get_xy(_xy) == piece and \
               not self.is_kingpinned(_xy, coord_after):
                res.append(_xy)

        x, y = coord_after
        res = []

        if piece.nature == PIECENATURE_PAWN:
            if movetype == MOVETYPE_SINGLE:
                if piece.color == COLOR_WHITE:
                    if y == 4 and self.get_xy((x, y+1)).is_empty():
                        add_to_res_if_rightpiece_notpinned((x, y+2), coord_after, piece)
                    else:
                        add_to_res_if_rightpiece_notpinned((x, y+1), coord_after, piece)
                else:  # piece.color == COLOR_BLACK
                    if y == 3 and self.get_xy((x, y-1)).is_empty():
                        add_to_res_if_rightpiece_notpinned((x, y-2), coord_after, piece)
                    else:
                        add_to_res_if_rightpiece_notpinned((x, y-1), coord_after, piece)
            elif movetype == MOVETYPE_CAPTURE:
                if piece.color == COLOR_WHITE:
                    add_to_res_if_rightpiece_notpinned((x+1, y+1), coord_after, piece)
                    add_to_res_if_rightpiece_notpinned((x-1, y+1), coord_after, piece)
                else:
                    add_to_res_if_rightpiece_notpinned((x+1, y-1), coord_after, piece)
                    add_to_res_if_rightpiece_notpinned((x-1, y-1), coord_after, piece)

        elif piece.nature == PIECENATURE_KNIGHT:
            add_to_res_if_rightpiece_notpinned((x-2, y-1), coord_after, piece)
            add_to_res_if_rightpiece_notpinned((x-2, y+1), coord_after, piece)
            add_to_res_if_rightpiece_notpinned((x+2, y-1), coord_after, piece)
            add_to_res_if_rightpiece_notpinned((x+2, y+1), coord_after, piece)
            add_to_res_if_rightpiece_notpinned((x-1, y-2), coord_after, piece)
            add_to_res_if_rightpiece_notpinned((x-1, y+2), coord_after, piece)
            add_to_res_if_rightpiece_notpinned((x+1, y-2), coord_after, piece)
            add_to_res_if_rightpiece_notpinned((x+1, y+2), coord_after, piece)
        elif piece.nature in (PIECENATURE_BISHOP,
                              PIECENATURE_ROOK,
                              PIECENATURE_QUEEN,
                              PIECENATURE_KING):
            deltamax = ChessBoard.moves_descr[piece.nature][0]
            for deltax, deltay in ChessBoard.moves_descr[piece.nature][1]:
                for delta in range(1, deltamax):
                    # ~~
                    if ChessBoard.xy_is_off_the_board((x+(deltax*delta), y+(deltay*delta))):
                        # we have to try another (deltax, deltay)
                        break
                    if self.is_empty_or_is_this_piece((x+(deltax*delta), y+(deltay*delta)), piece):
                        add_to_res_if_rightpiece_notpinned((x+(deltax*delta),
                                                            y+(deltay*delta)),
                                                           coord_after,
                                                           piece)
                        if self.get_xy((x+(deltax*delta), y+(deltay*delta))) == piece:
                            break
                    else:
                        break

        return res

    def who_attacks(self,
                    xy):
        """ChessBoard.who_attacks()"""

        def add_to_res_if_rightpiece(xy, piece):
            if not ChessBoard.xy_is_off_the_board(xy) and \
               self.get_xy(xy) == piece:
                res.append(xy)

        target = self.get_xy(xy)
        _color = invert_color(target.color)
        res = []

        x, y = xy

        # ---- do a pawn attack (x, y) ? --------------------------------------
        piece = ChessPiece(color=_color,
                           nature=PIECENATURE_PAWN)
        if target.color == COLOR_WHITE:
            add_to_res_if_rightpiece((x+1, y-1), piece)
            add_to_res_if_rightpiece((x-1, y-1), piece)
        else:
            add_to_res_if_rightpiece((x+1, y+1), piece)
            add_to_res_if_rightpiece((x-1, y+1), piece)

        # ---- do a knight attack (x, y) ? ------------------------------------
        piece = ChessPiece(color=_color,
                           nature=PIECENATURE_KNIGHT)
        add_to_res_if_rightpiece((x-2, y-1), piece)
        add_to_res_if_rightpiece((x-2, y+1), piece)
        add_to_res_if_rightpiece((x+2, y-1), piece)
        add_to_res_if_rightpiece((x+2, y+1), piece)
        add_to_res_if_rightpiece((x-1, y-2), piece)
        add_to_res_if_rightpiece((x-1, y+2), piece)
        add_to_res_if_rightpiece((x+1, y-2), piece)
        add_to_res_if_rightpiece((x+1, y+2), piece)

        # ---- do a bishop/rook/queen/king attack (x, y) ? --------------------
        for nature in (PIECENATURE_BISHOP,
                       PIECENATURE_ROOK,
                       PIECENATURE_QUEEN,
                       PIECENATURE_KING):
            piece = ChessPiece(color=_color,
                               nature=nature)
            deltamax = ChessBoard.moves_descr[nature][0]
            for deltax, deltay in ChessBoard.moves_descr[nature][1]:
                for delta in range(1, deltamax):
                    # ~~
                    if ChessBoard.xy_is_off_the_board((x+(deltax*delta), y+(deltay*delta))):
                        # we have to try another (deltax, deltay):
                        break
                    if self.is_empty_or_is_this_piece((x+(deltax*delta), y+(deltay*delta)), piece):
                        add_to_res_if_rightpiece((x+(deltax*delta),
                                                  y+(deltay*delta)),
                                                 piece)
                        if self.get_xy((x+(deltax*delta), y+(deltay*delta))) == piece:
                            break
                    else:
                        break

        return res

    @staticmethod
    def xy_is_off_the_board(xy):
        """ChessBoard.xy_is_off_the_board()"""
        return not ((0 <= xy[0] <= 7) and (0 <= xy[1] <= 7))


class ChessGame(Iaswn):
    """
        ChessGame class


        _______________________________________________________________________

        o  regex_pgn_tags
        o  regex_pgn_listofmoves
        o  regex_simplemove_algebraicnotation
        o  regex_simplemove_algebraicnotation2
        o  strcoord2coord / coord2strcoord

        o  (ChessGameTags)chessgame_tags
        o  (ChessBoard)board
        o  (ChessListOfMoves)listofmoves
        o  (ChessGameStatus)status
        o  (list of str)errors

        o  __eq__(self, other)
        o  __init__(self)
        o  __repr__(self)
        o  read_pgn(self, lines)
        o  read_pgn__doublemove(self, str_doublemove)
        o  read_pgn__listofmoves(self, src)
        o  read_pgn__simplemove(self, str_simplemove)
        o  write_pgn(self)
        o  write_pgn__listofmoves(self)
        o  write_pgn__simplemove(self, move)
    """
    # e.g. [Event "F/S Return Match"]
    regex_pgn_tags = re.compile(r'^\s*\[(?P<key>.+)\s+\"(?P<value>.+)\"\]$')

    regex_pgn_listofmoves = {
        'doublemovenumber': re.compile(r'[\d]+\.\s'),
        # Beware ! if you modify 'game_result', please update
        # ChessGameStatus.update_from_pgn_string().
        'game_result': re.compile(r'(?P<pre>.+)(?P<game_result>1\/2-1\/2|1\-0|0\-1|#|\*|\+\+)'),
        'en passant': re.compile(r'\s*e\.p\.'),
    }
    regex_simplemove_algebraicnotation = re.compile(
        r"^"
        r"(?P<str_piecenature>[KQNRB])?"
        r"(?P<str_intersymb>[x|\-])?"
        r"(?P<str_coord_x1>[a-h]|[1-8]|[a-h][1-8])"
        r"(=(?P<str_promotion>[KQNRB]))?"
        r"(?P<str_chess>\+)?"
        r"(?P<str_enpassant>e\.p\.)?"
        r"$")
    regex_simplemove_algebraicnotation2 = re.compile(
        r"^"
        r"(?P<str_piecenature>[KQNRB])?"
        r"(?P<str_coord_x0>[a-h][1-8]|[a-h]|[1-8])?"
        r"(?P<str_intersymb>[x|\-])?"
        r"(?P<str_coord_x1>[a-h][1-8]|[a-h]|[1-8])"
        r"(=(?P<str_promotion>[KQNRB]))?"
        r"(?P<str_chess>\+)?"
        r"(?P<str_enpassant>e\.p\.)?"
        r"$")

    strcoord2coord = {
        'a8': (0, 0), 'b8': (1, 0), 'c8': (2, 0), 'd8': (3, 0),
        'e8': (4, 0), 'f8': (5, 0), 'g8': (6, 0), 'h8': (7, 0),
        'a7': (0, 1), 'b7': (1, 1), 'c7': (2, 1), 'd7': (3, 1),
        'e7': (4, 1), 'f7': (5, 1), 'g7': (6, 1), 'h7': (7, 1),
        'a6': (0, 2), 'b6': (1, 2), 'c6': (2, 2), 'd6': (3, 2),
        'e6': (4, 2), 'f6': (5, 2), 'g6': (6, 2), 'h6': (7, 2),
        'a5': (0, 3), 'b5': (1, 3), 'c5': (2, 3), 'd5': (3, 3),
        'e5': (4, 3), 'f5': (5, 3), 'g5': (6, 3), 'h5': (7, 3),
        'a4': (0, 4), 'b4': (1, 4), 'c4': (2, 4), 'd4': (3, 4),
        'e4': (4, 4), 'f4': (5, 4), 'g4': (6, 4), 'h4': (7, 4),
        'a3': (0, 5), 'b3': (1, 5), 'c3': (2, 5), 'd3': (3, 5),
        'e3': (4, 5), 'f3': (5, 5), 'g3': (6, 5), 'h3': (7, 5),
        'a2': (0, 6), 'b2': (1, 6), 'c2': (2, 6), 'd2': (3, 6),
        'e2': (4, 6), 'f2': (5, 6), 'g2': (6, 6), 'h2': (7, 6),
        'a1': (0, 7), 'b1': (1, 7), 'c1': (2, 7), 'd1': (3, 7),
        'e1': (4, 7), 'f1': (5, 7), 'g1': (6, 7), 'h1': (7, 7),

        'a': (0, None), 'b': (1, None), 'c': (2, None), 'd': (3, None),
        'e': (4, None), 'f': (5, None), 'g': (6, None), 'h': (7, None),

        '1': (None, 7), '2': (None, 6), '3': (None, 5), '4': (None, 4),
        '5': (None, 3), '6': (None, 2), '7': (None, 1), '8': (None, 0),
        }
    coord2strcoord = {value: key for key, value in strcoord2coord.items()}

    def __eq__(self,
               other):
        """
            ChessGame.__eq__()

            Stricly speaking, not required to read/write PGN files but required
            to check that everything "works as expected" (see cwc validation).

            ___________________________________________________________________

            ARGUMENT: (ChessGame)other, the object compared to self

            RETURNED VALUE: (bool)True if other == self
        """
        return self.chessgame_tags == other.chessgame_tags and \
            self.board == other.board and \
            self.listofmoves == other.listofmoves and \
            self.status == other.status and \
            self.errors == other.errors

    def __init__(self):
        """ChessGame.__init__()"""
        Iaswn.__init__(self)
        self.chessgame_tags = ChessGameTags()
        self.board = ChessBoard()
        self.listofmoves = ChessListOfMoves()
        self.status = ChessGameStatus()

        self.errors = []

    def __repr__(self):
        """ChessGame.__repr__()"""
        res = f"{self.chessgame_tags=}; " \
            f"{self.board=}; {self.listofmoves=}; {self.status=}"
        if self.errors:
            res += "errors: "+str(self.errors)
        return res

    def read_pgn(self,
                 lines):
        """ChessGame.read_pgn()"""
        success = True

        self.listofmoves = ChessListOfMoves()
        try:
            str_listofmoves = []
            for _line in lines:
                line = _line.strip()

                if line:
                    regex_pgn__found = False
                    if res_regex_pgn_tags := re.search(ChessGame.regex_pgn_tags, line):
                        regex_pgn__found = True
                        self.chessgame_tags[res_regex_pgn_tags.group('key')] = \
                            res_regex_pgn_tags.group('value')

                    if regex_pgn__found is False:
                        str_listofmoves.append(line)

            str_listofmoves = " ".join(str_listofmoves)

            self.read_pgn__listofmoves(str_listofmoves)

        except ChessError as error:
            self.errors.append(error)
            success = False

        return success

    def read_pgn__doublemove(self,
                             str_doublemove):
        """ChessGame.read_pgn__doublemove()"""
        # " e.p." with space(s) must be rewritten "e.p." (without space)
        if re.search(ChessGame.regex_pgn_listofmoves['en passant'], str_doublemove):
            str_doublemove = re.sub(ChessGame.regex_pgn_listofmoves['en passant'],
                                    "e.p.",
                                    str_doublemove)
        # if <str_doublemove> is something like:
        #   "Kd5 Kxb3 0-1"
        # we have to attach the last part using the '_' character:
        #   "Kd5 Kxb3_0-1"
        if res__game_result := re.search(ChessGame.regex_pgn_listofmoves['game_result'],
                                         str_doublemove):
            self.status.update_from_pgn_string(status_string=res__game_result.group('game_result'),
                                               current_player=self.listofmoves.next_player)
            str_doublemove = re.sub(
                ChessGame.regex_pgn_listofmoves['game_result'],
                lambda match: match.group('pre').strip()+'_'+match.group('game_result').strip(),
                str_doublemove)

        if " " not in str_doublemove:
            # only one move in <str_doublemove> (as in 'Nc4')
            self.read_pgn__simplemove(str_doublemove)
        else:
            # normal case, two moves in <str_doublemove> (as in 'e4 e5')
            move1, move2 = str_doublemove.split(" ")
            self.read_pgn__simplemove(move1)
            self.read_pgn__simplemove(move2)

    def read_pgn__listofmoves(self,
                              src):
        """
            ChessGame.read_pgn__listofmoves()

            src: (list of str)listofmoves
        """
        doublemove_number = 0
        for _str_doublemove in re.split(ChessGame.regex_pgn_listofmoves['doublemovenumber'], src):
            str_doublemove = _str_doublemove.strip()
            if str_doublemove:
                doublemove_number += 1
                self.read_pgn__doublemove(str_doublemove)

    def read_pgn__simplemove(self,
                             str_simplemove):
        """simplemove: e6e4 // e6-e4 // e4"""

        # ---------------------------------------------------------------------
        # ---- let's remove from <str_simplemove> the game result. ------------
        # ---------------------------------------------------------------------
        str_game_result = None
        if res__game_result := re.search(ChessGame.regex_pgn_listofmoves['game_result'],
                                         str_simplemove):
            str_game_result = res__game_result.group('game_result')

            # let's remove the '_' character added by read_pgn__doublemove() before the game result:
            str_simplemove = str_simplemove.replace('_', '')
            # we can now remove the game result suffix (e.g. '1-0')
            self.status.update_from_pgn_string(status_string=res__game_result.group('game_result'),
                                               current_player=self.listofmoves.next_player)
            str_simplemove = re.sub(
                ChessGame.regex_pgn_listofmoves['game_result'],
                lambda match: match.group('pre').strip(),
                str_simplemove)

        str_simplemove = str_simplemove.strip()

        # ---------------------------------------------------------------------
        # ---- let's extract from str_simplemove the move details -------------
        # ---------------------------------------------------------------------

        # default values:
        who_plays = self.listofmoves.who_plays()
        movetype = MOVETYPE_SINGLE  # default value, may be changed.
        # if movetype is MOVETYPE_SINGLE, <piece1_coord_before> will be automatically
        # computed with the help of .which_piece_could_go_to():
        piece1_coord_before = None
        piece1_coord_after = None
        piece2_coord_before = None
        piece2_coord_after = None
        promotion = None  # if promotion, <promotion> will be the PIECENATURE_xxx constant
        enpassant = False

        # ---- let's try to initialize <piece1_coord_before> and --------------
        # ---- <piece1_coord_after> from <str_simplemove>. --------------------
        if str_simplemove == "O-O":
            movetype = MOVETYPE_CASTLING_OO
            # king
            piece1_coord_before = self.board.get_king_coord(color=who_plays)
            piece1_coord_after = piece1_coord_before[0]+2, piece1_coord_before[1]
            # rook:
            piece2_coord_before = piece1_coord_before[0]+3, piece1_coord_before[1]
            piece2_coord_after = piece1_coord_before[0]+1, piece1_coord_before[1]
        elif str_simplemove == "O-O-O":
            movetype = MOVETYPE_CASTLING_OOO
            # king
            piece1_coord_before = self.board.get_king_coord(color=who_plays)
            piece1_coord_after = piece1_coord_before[0]-2, piece1_coord_before[1]
            # rook:
            piece2_coord_before = piece1_coord_before[0]-4, piece1_coord_before[1]
            piece2_coord_after = piece1_coord_before[0]-1, piece1_coord_before[1]
        else:
            res_algebricnotation = {"str_piecenature": None,
                                    "str_coord_x0": None,
                                    "str_intersymb": None,
                                    "str_coord_x1": None,
                                    "str_promotion": None,
                                    "str_chess": None,
                                    "str_enpassant": None,
                                    }

            if _res := re.search(ChessGame.regex_simplemove_algebraicnotation,
                                 str_simplemove):
                res_algebricnotation["str_piecenature"] = _res.group("str_piecenature")
                res_algebricnotation["str_intersymb"] = _res.group("str_intersymb")
                res_algebricnotation["str_coord_x1"] = _res.group("str_coord_x1")
                res_algebricnotation["str_promotion"] = _res.group("str_promotion")
                res_algebricnotation["str_chess"] = _res.group("str_chess")
                res_algebricnotation["str_enpassant"] = _res.group("str_enpassant")
            elif _res := re.search(ChessGame.regex_simplemove_algebraicnotation2,
                                   str_simplemove):
                res_algebricnotation["str_piecenature"] = _res.group("str_piecenature")
                res_algebricnotation["str_coord_x0"] = _res.group("str_coord_x0")
                res_algebricnotation["str_intersymb"] = _res.group("str_intersymb")
                res_algebricnotation["str_coord_x1"] = _res.group("str_coord_x1")
                res_algebricnotation["str_promotion"] = _res.group("str_promotion")
                res_algebricnotation["str_chess"] = _res.group("str_chess")
                res_algebricnotation["str_enpassant"] = _res.group("str_enpassant")
            else:
                raise ChessError(
                    f"Ill-formed or unknown string format for str_simplemove='{str_simplemove}'.")

            if res_algebricnotation["str_intersymb"] == "x":
                movetype = MOVETYPE_CAPTURE

            if res_algebricnotation["str_piecenature"] is not None:
                # str_piecenature: 'Q' / 'K' / 'R' / 'B' / 'N'
                piece1_piecenature = \
                    ALGEBRICNOTATION2PIECENATURE[res_algebricnotation["str_piecenature"]]
            else:
                # no piecenature: it's a pawn
                piece1_piecenature = PIECENATURE_PAWN

            if res_algebricnotation["str_coord_x0"] is None:  # coord_x0 wasn't given (e.g. 'Rxb4')
                piece1_coord_after = ChessGame.strcoord2coord[res_algebricnotation["str_coord_x1"]]
            else:  # coord_x0 was given, at least partially (e.g. 'cxb5', 'Rbxb4')
                # piece1_coord_before > maybe partial (e.g. 'Rbb4')
                piece1_coord_before = ChessGame.strcoord2coord[res_algebricnotation["str_coord_x0"]]
                piece1_coord_after = ChessGame.strcoord2coord[res_algebricnotation["str_coord_x1"]]

                if piece1_coord_before[1] is None:
                    # <piece1_coord_before> is only partially initialized: we only have the column,
                    # as in 'Qab2', 'cxb5'
                    for _x, _y in self.board.which_piece_could_go_to(
                            piece=ChessPiece(nature=piece1_piecenature,
                                             color=self.listofmoves.next_player),
                            coord_after=piece1_coord_after,
                            movetype=movetype):
                        if _x == piece1_coord_before[0]:
                            # a tuple would not be writable:
                            piece1_coord_before = list(piece1_coord_before)
                            piece1_coord_before[1] = _y  # we found the row
                            piece1_coord_before = tuple(piece1_coord_before)
                            #
                            # To optimize, you may want to consider the first solution is the right
                            # one and therefore add 'break'; without this 'break' statement you may
                            # find errors in the list of moves.
                            #
                            # break

                elif piece1_coord_before[0] is None:
                    # <piece1_coord_before> is only partially initialized: we only have the row,
                    # as in 'Q2b2'.
                    for _x, _y in self.board.which_piece_could_go_to(
                            piece=ChessPiece(nature=piece1_piecenature,
                                             color=self.listofmoves.next_player),
                            coord_after=piece1_coord_after,
                            movetype=movetype):
                        if _y == piece1_coord_before[1]:
                            # a tuple would not be writable:
                            piece1_coord_before = list(piece1_coord_before)
                            piece1_coord_before[0] = _x  # we found the column
                            piece1_coord_before = tuple(piece1_coord_before)
                            #
                            # To optimize, you may want to consider the first solution is the right
                            # one and therefore add 'break'; without this 'break' statement you may
                            # find errors in the list of moves.
                            #
                            # break

            if res_algebricnotation["str_promotion"]:
                promotion = ALGEBRICNOTATION2PIECENATURE[res_algebricnotation["str_promotion"]]

            if piece1_coord_after[0] is None or piece1_coord_after[1] is None:
                raise ChessError(f"Can't understand this (simple) move: {str_simplemove} .")

        # ---- special case: piece1_coord_before has not yet been initialized -
        # It happens if <str_simplemove> contains an extra character to remove
        # any ambiguity.
        if piece1_coord_before is None:
            _possi = self.board.which_piece_could_go_to(
                piece=ChessPiece(nature=piece1_piecenature,
                                 color=self.listofmoves.next_player),
                coord_after=piece1_coord_after,
                movetype=movetype)
            if not _possi:
                raise ChessError(
                    f"Can't interpret (simple) move '{str_simplemove}' for the current board. "
                    f"No legal move matches this string. "
                    f"self={repr(self)}")
            if len(_possi) > 1:
                raise ChessError(
                    f"Anomaly: too many possibilities, the move '{str_simplemove}' is ambiguous. "
                    f"Current situation is: {repr(self)}")
            # [0] since there is only ONE POSSIBILITY here: len(_possi) is 1.
            piece1_coord_before = _possi[0]

        # ---- en passant ? ---------------------------------------------------
        enpassant = movetype == MOVETYPE_CAPTURE and \
            self.board.get_xy((piece1_coord_after[0], piece1_coord_after[1])).is_empty()

        # ---- new_move, update of .listofmoves and .board ----------------
        new_move = ChessMove(movetype=movetype,
                             beforeafter_coord_piece1=(piece1_coord_before, piece1_coord_after),
                             beforeafter_coord_piece2=(piece2_coord_before, piece2_coord_after),
                             promotion=promotion,
                             enpassant=enpassant,
                             str_game_result=str_game_result)
        self.listofmoves.add_move(new_move)
        self.board.update_by_playing_a_move(new_move)

    def write_pgn(self):
        """ChessGame.write_pgn()"""
        res = []

        for key, value in self.chessgame_tags.items():
            res.append(f'[{key} "{value}"]')

        res.append("")

        res.extend(self.write_pgn__listofmoves())

        return res

    def write_pgn__listofmoves(self):
        """ChessGame.write_pgn__listofmoves()"""
        res = []

        self.board.set_startpos()

        doublemove_str = ""
        for simplemove_index, (doublemove_index, _, simplemove) in enumerate(self.listofmoves):
            if simplemove_index % 2 == 0:
                # white plays:
                doublemove_str = str(doublemove_index)+". " + self.write_pgn__simplemove(simplemove)
            else:
                # black plays:
                doublemove_str += " " + self.write_pgn__simplemove(simplemove)
                res.append(doublemove_str)
                doublemove_str = ""

        if doublemove_str:
            res.append(doublemove_str)

        return res

    def write_pgn__simplemove(self,
                              move):
        """ChessGame.write_pgn__simplemove()"""
        piece1 = self.board.get_xy(move.beforeafter_coord_piece1[0])

        _possi = None
        if move.movetype in (MOVETYPE_SINGLE, MOVETYPE_CAPTURE):
            _possi = self.board.which_piece_could_go_to(piece1,
                                                        move.beforeafter_coord_piece1[1],
                                                        move.movetype)

        self.board.update_by_playing_a_move(move)

        if move.movetype == MOVETYPE_CASTLING_OO:
            return "O-O"

        if move.movetype == MOVETYPE_CASTLING_OOO:
            return "O-O-O"

        res = ""
        if piece1.nature != PIECENATURE_PAWN:
            res += PIECENATURE2ALGEBRICNOTATION[piece1.nature]

        if len(_possi) == 1:
            if piece1.nature == PIECENATURE_PAWN and move.movetype == MOVETYPE_CAPTURE:
                # 'd' in 'dxe5'
                res += ChessGame.coord2strcoord[(move.beforeafter_coord_piece1[0][0], None)]
            else:
                pass
        else:
            if piece1.nature == PIECENATURE_PAWN:
                # 'd' in 'dxe5'
                res += ChessGame.coord2strcoord[(move.beforeafter_coord_piece1[0][0], None)]
            else:
                # The ambiguity must be resolved by specifying only the column ('a', 'b', ...) or,
                # if the column does not resolve the ambiguity, only the row ('1', '2', ...);
                # otherwise, the column and row must be written.
                #
                # E.g. if beforeafter_coord_piece1[0] is equal to (2, 0) (==[c8]),
                # we count the number of times (int)2 appears in _possi: it must
                # appear at last one time(*); if it's the case, there's no ambiguity to add only
                # the column; it (int)2 appears more than one time, the ambiguity is not resolved.
                #
                # (*) Why at least one time ? Because among _possiilites lies the right move, the
                #     one that has been chosen. If this possibility is ambiguous, other moves will
                #     add their column/row.
                if tuple(_x for _x, _y in _possi).count(move.beforeafter_coord_piece1[0][0]) == 1:
                    res += ChessGame.coord2strcoord[(move.beforeafter_coord_piece1[0][0], None)]
                elif tuple(_y for _x, _y in _possi).count(move.beforeafter_coord_piece1[0][1]) == 1:
                    res += ChessGame.coord2strcoord[(None, move.beforeafter_coord_piece1[0][1])]
                else:
                    res += ChessGame.coord2strcoord[move.beforeafter_coord_piece1[0]]

        if move.movetype == MOVETYPE_SINGLE:
            res += ChessGame.coord2strcoord[move.beforeafter_coord_piece1[1]]
        elif move.movetype == MOVETYPE_CAPTURE:
            res += "x"+ChessGame.coord2strcoord[move.beforeafter_coord_piece1[1]]

        if move.promotion:
            res += "="+PIECENATURE2ALGEBRICNOTATION[move.promotion]

        if move.enpassant:
            res += " e.p."

        if move.str_game_result:
            res += " " + move.str_game_result

        return res


class ChessGames(list, Iaswn):
    """
        ChessGames class


        _______________________________________________________________________

        o  __eq__(self, other)
        o  read_pgn(self, pgnfilename)
        o  write_pgn(self, pgnfilename)
    """
    def __eq__(self,
               other):
        """
            ChessGames.__eq__()

            Stricly speaking, not required to read/write PGN files but required
            to check that everything "works as expected" (see cwc validation).

            ___________________________________________________________________

            ARGUMENT: (ChessGames)other, the object compared to self

            RETURNED VALUE: (bool)True if other == self
        """
        return list.__eq__(self, other)

    def read_pgn(self,
                 src):
        """ChessGames.read_pgn()"""
        success = True

        inside_header = False
        buff = []

        for _line in src:
            line = _line.strip()

            if line:
                if inside_header:
                    if re.search(ChessGame.regex_pgn_tags, line) and not line.startswith("1. "):
                        # we're still in the header.
                        pass
                    else:
                        # we're not in the header anymore.
                        inside_header = False
                else:
                    if re.search(ChessGame.regex_pgn_tags, line) and not line.startswith("1. "):
                        # we weren't in the header but <line> is a header line.
                        if buff:
                            game = ChessGame()
                            success = success and game.read_pgn(buff)
                            self.append(game)
                        buff = []
                        inside_header = True
                    else:
                        # we're still reading data lines that are not in a header.
                        if line.startswith("1. "):
                            if buff:
                                game = ChessGame()
                                success = success and game.read_pgn(buff)
                                self.append(game)
                            buff = []

                buff.append(line)

        if buff:
            game = ChessGame()
            success = success and game.read_pgn(buff)
            self.append(game)

        return success

    def write_pgn(self,
                  dest):
        """ChessGames.write_pgn()"""
        for game in self:
            for line in game.write_pgn():
                dest.write(line+"\n")
            dest.write("\n")
